#     Copyright 2025, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""This contains the tuning of the compilers towards defined goals."""

import os
import re

from SCons.Script import (  # pylint: disable=I0021,import-error
    ARGUMENTS,
    Environment,
    GetOption,
)

from nuitka.Tracing import scons_details_logger, scons_logger
from nuitka.utils.Download import getCachedDownloadedMinGW64
from nuitka.utils.FileOperations import (
    getNormalizedPathJoin,
    getReportPath,
    openTextFile,
    putTextFileContents,
)
from nuitka.utils.PrivatePipSpace import getZigBinaryPath
from nuitka.utils.Utils import (
    isAIX,
    isFedoraBasedLinux,
    isLinux,
    isMacOS,
    isPosixWindows,
    isWin32Windows,
)

from .SconsHacks import getEnhancedToolDetect, myDetectVersion
from .SconsProgress import enableSconsProgressBar
from .SconsUtils import (
    addBinaryBlobSection,
    addToPATH,
    createEnvironment,
    decideArchMismatch,
    enableFlagSettings,
    getArgumentBool,
    getArgumentDefaulted,
    getArgumentInt,
    getArgumentList,
    getArgumentRequired,
    getExecutablePath,
    getLinkerArch,
    getMsvcVersion,
    getMsvcVersionString,
    initScons,
    isClangName,
    isGccName,
    isZigName,
    raiseNoCompilerFoundErrorExit,
    setEnvironmentVariable,
    setupScons,
)

# spell-checker: ignore LIBPATH,CPPDEFINES,CPPPATH,CXXVERSION,CCFLAGS,LINKFLAGS,CXXFLAGS
# spell-checker: ignore -flto,-fpartial-inlining,-freorder-functions,-defsym,-fprofile
# spell-checker: ignore -fwrapv,-Wunused,fcompare,-ftrack,-fvisibility,-municode
# spell-checker: ignore -feliminate,noexecstack,implib,bexpall,blibpath
# spell-checker: ignore LTCG,GENPROFILE,USEPROFILE,CGTHREADS,CCVERSION


def _detectWindowsSDK(env):
    # Caching
    if hasattr(env, "windows_sdk_version"):
        return env.windows_sdk_version

    # Check if there is a Windows SDK installed.
    if "WindowsSDKVersion" not in env:
        if "WindowsSDKVersion" in os.environ:
            windows_sdk_version = os.environ["WindowsSDKVersion"].rstrip("\\")
        else:
            windows_sdk_version = None
    else:
        windows_sdk_version = env["WindowsSDKVersion"]

    if windows_sdk_version:
        scons_details_logger.info("Using Windows SDK '%s'." % windows_sdk_version)

        env.windows_sdk_version = tuple(int(x) for x in windows_sdk_version.split("."))
    else:
        scons_logger.warning(
            """\
Windows SDK must be installed in Visual Studio for it to \
be usable with Nuitka. Use the Visual Studio installer for \
adding it."""
        )

        env.windows_sdk_version = None

    return env.windows_sdk_version


_windows_sdk_c11_mode_min_version = (10, 0, 19041, 0)


def _enableC11Settings(env):
    """Decide if C11 mode can be used and enable the C compile flags for it.

    Args:
        env - scons environment with compiler information

    Returns:
        bool - c11_mode flag
    """

    # Lots of cases to deal with

    if "force-c11-mode" in env.experimental_flags:
        c11_mode = True
    elif "force-cpp-mode" in env.experimental_flags:
        c11_mode = False
    elif env.clangcl_mode:
        c11_mode = True
    elif (
        env.msvc_mode
        and env.windows_sdk_version >= _windows_sdk_c11_mode_min_version
        and getMsvcVersion(env) >= (14, 3)
    ):
        c11_mode = True
    elif env.clang_mode or env.zig_mode:
        c11_mode = True
    elif env.gcc_mode and env.gcc_version >= (5,):
        c11_mode = True
    else:
        c11_mode = False

    if c11_mode:
        if env.gcc_mode or env.zig_mode:
            env.Append(CFLAGS=["-std=c11"])
        elif env.msvc_mode:
            env.Append(CFLAGS=["/std:c11"])

    if env.msvc_mode and c11_mode:
        # Windows SDK shows this even in non-debug mode in C11 mode.
        env.Append(CCFLAGS=["/wd5105", "/wd4391"])

    if not c11_mode:
        env.Append(CPPDEFINES=["_NUITKA_NON_C11_MODE"])

    scons_details_logger.info("Using C11 mode: %s" % c11_mode)

    env.c11_mode = c11_mode


def _enableLtoSettings(
    env,
    lto_mode,
    pgo_mode,
):
    # This is driven by branches on purpose and pylint: disable=too-many-branches,too-many-statements

    orig_lto_mode = lto_mode

    if lto_mode == "no":
        lto_mode = False
        reason = "disabled"
    elif lto_mode == "yes":
        lto_mode = True
        reason = "enabled"
    elif pgo_mode in ("use", "generate"):
        lto_mode = True
        reason = "PGO implies LTO"
    elif env.msvc_mode and getMsvcVersion(env) >= (14,):
        lto_mode = True
        reason = "known to be supported"
    elif env.monolithpy:
        lto_mode = True
        reason = "known to be supported (MonolithPy)"
    elif env.fedora_python:
        lto_mode = True
        reason = "known to be supported (Fedora Python)"
    elif env.arch_python:
        lto_mode = True
        reason = "known to be supported (Arch Python)"
    elif (
        env.debian_python
        and (env.gcc_mode or env.zig_mode)
        and not env.clang_mode
        and env.gcc_version >= (6,)
    ):
        lto_mode = True
        reason = "known to be supported (Debian)"
    elif env.gcc_mode and "gnu-cc" in env.the_cc_name and env.anaconda_python:
        lto_mode = False
        reason = "known to be not supported (CondaCC)"
    elif isMacOS() and (env.gcc_mode or env.zig_mode) and env.clang_mode:
        if env.zig_mode:
            lto_mode = False
            reason = "known to be not supported (zig on macOS)"
        elif env.debugger_mode:
            lto_mode = False
            reason = "must be disabled to see line numbers (macOS clang)"
        else:
            lto_mode = True
            reason = "known to be supported (macOS clang)"
    elif env.mingw_mode and env.clang_mode:
        lto_mode = False
        reason = "known to not be supported (new MinGW64 Clang)"
    elif env.gcc_mode and env.mingw_mode and env.gcc_version >= (11, 2):
        lto_mode = True
        reason = "known to be supported (new MinGW64)"
    else:
        lto_mode = False
        reason = "not known to be supported"

    # Do not default to LTO for large compilations, unless asked for it.
    compiled_module_count_threshold = 250
    if (
        orig_lto_mode == "auto"
        and lto_mode
        and env.compiled_module_count > compiled_module_count_threshold
        and not env.monolithpy
    ):
        lto_mode = False
        reason = (
            "might to be too slow %s (>= %d threshold), force it with '--lto=yes'"
            % (
                env.compiled_module_count,
                compiled_module_count_threshold,
            )
        )

    if lto_mode and env.gcc_mode and not env.clang_mode and env.gcc_version < (4, 6):
        scons_logger.warning(
            """\
The gcc compiler %s (version %s) doesn't have the sufficient \
version for lto mode (>= 4.6). Disabled."""
            % (env["CXX"], env["CXXVERSION"])
        )

        lto_mode = False
        reason = "gcc 4.6 is doesn't have good enough LTO support"

    if (
        lto_mode
        and (env.gcc_mode or env.zig_mode)
        and not env.clang_mode
        and not isWin32Windows()
        and not isMacOS()
        and getExecutablePath("make", env=env) is None
    ):
        scons_logger.warning(
            """\
The gcc compiler for LTO mode requires 'make' to be installed \
for parallel linking to be used, compilation might be a lot \
slower without it.
"""
        )

    if (env.gcc_mode or env.zig_mode) and lto_mode:
        if env.clang_mode:
            if "thin-lto" not in env.experimental_flags:
                env.Append(CCFLAGS=["-flto"])
                env.Append(LINKFLAGS=["-flto"])
            else:
                env.Append(CCFLAGS=["-flto=thin"])
                env.Append(LINKFLAGS=["-flto=thin"])
        else:
            env.Append(CCFLAGS=["-flto=%d" % env.job_count])
            env.Append(LINKFLAGS=["-flto=%d" % env.job_count])

            env.Append(CCFLAGS=["-fuse-linker-plugin", "-fno-fat-lto-objects"])
            env.Append(LINKFLAGS=["-fuse-linker-plugin"])

            # Need to tell the linker these things are OK.
            env.Append(LINKFLAGS=["-fpartial-inlining", "-freorder-functions"])

            if env.mingw_mode and "MAKE" not in os.environ:
                setEnvironmentVariable(env, "MAKE", "mingw32-make.exe")

    # Tell compiler to use link time optimization for MSVC
    if env.msvc_mode and lto_mode:
        if not env.clangcl_mode:
            env.Append(CCFLAGS=["/GL"])

            if getMsvcVersion(env) >= (14, 3):
                env.Append(LINKFLAGS=["/CGTHREADS:%d" % env.job_count])

        env.Append(LINKFLAGS=["/LTCG"])

    if orig_lto_mode == "auto":
        scons_details_logger.info(
            "LTO mode auto was resolved to mode: '%s' (%s)."
            % ("yes" if lto_mode else "no", reason)
        )

    env.lto_mode = lto_mode
    env.orig_lto_mode = orig_lto_mode
    env.pgo_mode = pgo_mode

    # PGO configuration
    _enablePgoSettings(env)


_python311_min_msvc_version = (14, 3)


def checkWindowsCompilerFound(
    env,
    target_arch,
    mingw_mode,
    clang_mode,
    msvc_version,
    download_ok,
    assume_yes_for_downloads,
    experimental_flags,
):
    """Remove compiler of wrong arch or too old gcc and replace with downloaded winlibs gcc."""
    # Many cases to deal with, pylint: disable=too-many-branches,too-many-statements

    if os.name != "nt" or isZigName(env["CC"]):
        return env

    # On Windows, in case MSVC was not found and not previously forced, use the
    # winlibs MinGW64 or zig as a download, and use it as a fallback.
    compiler_path = getExecutablePath(env["CC"], env=env)

    scons_details_logger.info(
        "Checking usability of binary '%s' from environment '%s'."
        % (compiler_path, env["CC"])
    )

    # On MSYS2, cannot use the POSIX compiler, drop that even before we check arches, since that
    # will of course still match.
    if (
        env.msys2_mingw_python
        and compiler_path is not None
        and compiler_path.endswith("/usr/bin/gcc.exe")
    ):
        scons_details_logger.info(
            "MSYS2 gcc found, but not used as it'S considered unusable."
        )
        compiler_path = None

    if compiler_path is not None:
        the_cc_name = os.path.basename(compiler_path)

        # The MSVC can only be used with an Windows SDK installed, and for 3.11 we need it
        # to be a least a minimum version.
        if (
            not isGccName(the_cc_name)
            and not isClangName(the_cc_name)
            and (
                _detectWindowsSDK(env) is None
                or (
                    env.python_version is not None
                    and env.python_version >= (3, 11)
                    and _detectWindowsSDK(env) < _windows_sdk_c11_mode_min_version
                )
            )
        ):
            # This will trigger using it to use our own gcc in branch below.
            compiler_path = None
            env["CC"] = None

    if compiler_path is not None:
        the_cc_name = os.path.basename(compiler_path)

        if clang_mode and not isClangName(the_cc_name):
            # This will trigger using it to use our own clang in branch below.
            compiler_path = None
            env["CC"] = None

    # Drop wrong arch compiler, most often found by scans. There might be wrong gcc or cl on the PATH.
    if compiler_path is not None:
        the_cc_name = os.path.basename(compiler_path)

        decision, linker_arch, compiler_arch = decideArchMismatch(
            target_arch=target_arch,
            the_cc_name=the_cc_name,
            compiler_path=compiler_path,
        )

        if decision:
            scons_logger.info(
                """\
Mismatch between Python binary ('%s' -> '%s') and \
C compiler ('%s' -> '%s') arches, that compiler is ignored!"""
                % (
                    os.environ["NUITKA_PYTHON_EXE_PATH"],
                    linker_arch,
                    compiler_path,
                    compiler_arch,
                )
            )

            # This will trigger using it to use our own gcc in branch below.
            compiler_path = None
            env["CC"] = None

    if compiler_path is not None and msvc_version is not None:
        if msvc_version == "latest":
            scons_logger.info(
                "MSVC version resolved to %s." % getMsvcVersionString(env)
            )
        # Requested a specific MSVC version, check if that worked.
        elif msvc_version != getMsvcVersionString(env):
            scons_logger.info(
                "Failed to find requested MSVC version ('%s' != '%s')."
                % (msvc_version, getMsvcVersionString(env))
            )

            # This will trigger error exit in branch below.
            compiler_path = None
            env["CC"] = None

    # For Python3.11+ we need to check Windows SDK version.
    if compiler_path is not None:
        the_cc_name = os.path.basename(compiler_path)

        # The MSVC can only be used with an Windows SDK installed, and for 3.11 we need it
        # to be a least a minimum version.
        if (
            # This is actually OK to use like this, pylint: disable=bad-chained-comparison
            not isGccName(the_cc_name)
            and None is not env.python_version >= (3, 11)
            and getMsvcVersion(env) < _python311_min_msvc_version
        ):
            scons_logger.info(
                """\
For Python version %s MSVC %s or later is required, not %s which is too old."""
                % (
                    ".".join(str(d) for d in env.python_version),
                    ".".join(str(d) for d in _python311_min_msvc_version),
                    getMsvcVersionString(env),
                )
            )

            # This will trigger error exit in branch below.
            compiler_path = None
            env["CC"] = None

    if compiler_path is not None:
        the_cc_name = os.path.basename(compiler_path)

        if isGccName(the_cc_name):
            if "force-accept-windows-gcc" not in experimental_flags:
                scons_logger.info(
                    "Non downloaded winlibs-gcc '%s' is being ignored, Nuitka is very dependent on the precise one."
                    % (compiler_path,)
                )

                # This also will trigger using it to use our own gcc in branch below.
                compiler_path = None
                env["CC"] = None

    # This will download "gcc.exe" (or "zig.exe") when all others have been
    # rejected and MSVC is not enforced.

    if compiler_path is None and msvc_version is None:
        if env.python_version < (3, 13):
            scons_details_logger.info(
                "No usable C compiler, attempt fallback to winlibs gcc."
            )

            compiler_path = getCachedDownloadedMinGW64(
                target_arch=target_arch,
                assume_yes_for_downloads=assume_yes_for_downloads,
                download_ok=download_ok,
                experimental="winlibs-new" in experimental_flags,
            )
        elif target_arch != "x86" and not mingw_mode:
            scons_details_logger.info("No usable C compiler, attempt fallback to zig.")

            # This will download "zig.exe" when all others have been rejected
            # and MSVC is not enforced.
            compiler_path = getZigBinaryPath(
                logger=scons_logger,
                assume_yes_for_downloads=env.assume_yes_for_downloads,
                reject_message=None,
            )

        if compiler_path is not None:
            addToPATH(env=env, dirname=os.path.dirname(compiler_path), prefix=True)

            env = createEnvironment(
                mingw_mode=True,
                msvc_version=None,
                clang_mode=False,
                clangcl_mode=False,
                target_arch=target_arch,
                consider_environ_variables=False,
                assume_yes_for_downloads=env.assume_yes_for_downloads,
                source_dir=env.source_dir,
            )

            if clang_mode:
                env["CC"] = os.path.join(os.path.dirname(compiler_path), "clang.exe")
        else:
            raiseNoCompilerFoundErrorExit()

    return env


def createEnvironmentAndCheckCompiler(
    mingw_mode,
    msvc_version,
    clang_mode,
    clangcl_mode,
    target_arch,
    consider_environ_variables,
    assume_yes_for_downloads,
    download_ok,
    experimental_flags,
    source_dir,
):
    anaconda_python = getArgumentBool("anaconda_python", False)

    if isLinux() and anaconda_python:
        python_prefix = getArgumentRequired("python_prefix")
        addToPATH(None, getNormalizedPathJoin(python_prefix, "bin"), prefix=True)

    env = createEnvironment(
        mingw_mode=mingw_mode,
        msvc_version=msvc_version,
        clang_mode=clang_mode,
        clangcl_mode=clangcl_mode,
        target_arch=target_arch,
        consider_environ_variables=consider_environ_variables,
        assume_yes_for_downloads=assume_yes_for_downloads,
        source_dir=source_dir,
    )

    env = checkWindowsCompilerFound(
        env=env,
        target_arch=target_arch,
        mingw_mode=mingw_mode,
        clang_mode=clang_mode,
        msvc_version=msvc_version,
        download_ok=download_ok,
        assume_yes_for_downloads=assume_yes_for_downloads,
        experimental_flags=experimental_flags,
    )

    env.the_compiler = env["CC"] or env["CXX"]
    env.the_cc_name = os.path.normcase(os.path.basename(env.the_compiler))

    # Requested or user provided, detect if it's clang even from environment
    if isClangName(env.the_cc_name):
        clang_mode = True
        env["CCVERSION"] = None

    # We consider clang to be a form of gcc for the most things, they strive to
    # be compatible.
    env.zig_mode = isZigName(env.the_cc_name)
    env.gcc_mode = (
        isGccName(env.the_cc_name)
        or clang_mode
        or (env.zig_mode and not os.name == "nt")
    )
    env.clang_mode = clang_mode

    # Only use MSVC if not already clear, we are using MinGW.
    env.msvc_mode = os.name == "nt" and not env.gcc_mode and not env.zig_mode
    env.mingw_mode = os.name == "nt" and env.gcc_mode and not env.zig_mode
    env.clangcl_mode = clangcl_mode

    # For Python3.13, we need to enforce it for now to use MSVC
    if (
        os.name == "nt"
        and not env.msvc_mode
        and not env.zig_mode
        and env.python_version >= (3, 13)
    ):
        return scons_logger.sysexit(
            """\
Sorry, MinGW64 is not currently supported with Python 3.13,
due to differences in layout internal structures of Python.

Newer Nuitka will work eventually to solve this. Either use
Python 3.12 or option "--zig" (another Free Software compiler)
as a workaround for now and wait for updates of Nuitka to add
MinGW64 support back. Alternatively you can also install MSVC
as described in the documentation.""",
            env=env,
        )

    # Detect versions, ensuring they are always set to a tuple, so comparisons
    # do not fail with TypeError.
    if env.zig_mode:
        env.zig_version = myDetectVersion(env, env.the_compiler)

        if env.zig_version is None:
            scons_logger.sysexit(
                "Error, failed to detect Zig version of backend compiler '%s'."
                % env.the_compiler
            )

        env.clang_version = None
        env.gcc_version = None
    elif env.clang_mode:
        env.clang_version = myDetectVersion(env, env.the_compiler)

        if env.clang_version is None:
            scons_logger.sysexit(
                "Error, failed to detect Clang version of backend compiler '%s'."
                % env.the_compiler
            )

        env.zig_version = None
        env.gcc_version = None
    elif env.gcc_mode:
        env.gcc_version = myDetectVersion(env, env.the_compiler)

        if env.gcc_version is None:
            scons_logger.sysexit(
                "Error, failed to detect GCC version of backend compiler '%s'."
                % env.the_compiler
            )

        env.clang_version = None
        env.zig_version = None
    else:
        env.clang_version = None
        env.zig_version = None
        env.gcc_version = None

    return env


def decideConstantsBlobResourceMode(env):
    # This is a complicated decision with a lot of cases, as there are many
    # compiler, mode, OS and their versions related decisions.
    # pylint: disable=too-many-branches

    if "NUITKA_RESOURCE_MODE" in os.environ:
        resource_mode = os.environ["NUITKA_RESOURCE_MODE"]
        reason = "user provided"
    elif env.zig_mode:
        resource_mode = "c23_embed"
        reason = "default for zig"
    elif isWin32Windows():
        if env.clangcl_mode and getMsvcVersion(env) >= (14, 3):
            resource_mode = "c23_embed"
            reason = "default for ClangCL"
        else:
            resource_mode = "win_resource"
            reason = "default for Windows"
    elif isPosixWindows():
        resource_mode = "linker"
        reason = "default MSYS2 Posix"
    elif isMacOS():
        # TODO: The macOS has no Clang19 yet, once it does, we could also
        # consider using "c23_embed" for it too.
        resource_mode = "mac_section"
        reason = "default for macOS"
    elif env.gcc_mode and env.clang_mode and env.clang_version >= (19,):
        resource_mode = "c23_embed"
        reason = "default for newer clang"
    elif env.gcc_mode and not env.clang_mode and env.gcc_version >= (15,):
        resource_mode = "c23_embed"
        reason = "default for newer gcc"
    elif isAIX():
        resource_mode = "code"
        reason = "AIX is not compatible with incbin"
    elif env.lto_mode and env.gcc_mode and not env.clang_mode:
        if env.module_mode:
            resource_mode = "code"
        else:
            resource_mode = "linker"

        reason = "default for lto gcc with '--lto' and bugs for incbin"
    else:
        # All is done already, this is for most platforms.
        resource_mode = "incbin"
        reason = "default"

    assert resource_mode in (
        "incbin",
        "linker",
        "win_resource",
        "mac_section",
        "code",
        "c23_embed",
    ), resource_mode

    return resource_mode, reason


def addConstantBlobFile(env, blob_filename, resource_desc):
    resource_mode, reason = resource_desc

    assert blob_filename.endswith(".bin"), blob_filename

    scons_details_logger.info(
        "Using resource mode: '%s' (%s)." % (resource_mode, reason)
    )

    if resource_mode == "win_resource":
        # On Windows constants can be accessed as a resource by Nuitka at run time afterwards.
        env.Append(CPPDEFINES=["_NUITKA_CONSTANTS_FROM_RESOURCE"])
    elif resource_mode == "mac_section":
        env.Append(CPPDEFINES=["_NUITKA_CONSTANTS_FROM_MACOS_SECTION"])

        addBinaryBlobSection(
            env=env,
            blob_filename=blob_filename,
            section_name=os.path.basename(blob_filename)[:-4].lstrip("_"),
        )
    elif resource_mode == "incbin":
        env.Append(CPPDEFINES=["_NUITKA_CONSTANTS_FROM_INCBIN"])

        constants_generated_filename = os.path.join(
            env.source_dir, "__constants_data.c"
        )

        putTextFileContents(
            constants_generated_filename,
            contents=r"""
#define INCBIN_PREFIX
#define INCBIN_STYLE INCBIN_STYLE_SNAKE
#define INCBIN_LOCAL
#ifdef _NUITKA_EXPERIMENTAL_WRITEABLE_CONSTANTS
#define INCBIN_OUTPUT_SECTION ".data"
#endif

#include "nuitka/incbin.h"

INCBIN(constant_bin, "%(blob_filename)s");

unsigned char const *getConstantsBlobData(void) {
    return constant_bin_data;
}
"""
            % {"blob_filename": blob_filename},
        )

    elif resource_mode == "linker":
        # Indicate "linker" resource mode.
        env.Append(CPPDEFINES=["_NUITKA_CONSTANTS_FROM_LINKER"])

        # For MinGW the symbol name to be used is more low level.
        constant_bin_link_name = "constant_bin_data"
        if env.mingw_mode:
            constant_bin_link_name = "_" + constant_bin_link_name

        # At least some gcc versions need this path to contain a path element,
        # but it's normalized when it gets to here.
        if env.source_dir == ".":
            blob_filename = "./%s" % blob_filename

        env.Append(
            LINKFLAGS=[
                "-Wl,-b",
                "-Wl,binary",
                "-Wl,%s" % blob_filename,
                "-Wl,-b",
                "-Wl,%s"
                % getLinkerArch(
                    target_arch=env.target_arch,
                    mingw_mode=env.mingw_mode or isPosixWindows(),
                ),
                "-Wl,-defsym",
                "-Wl,%s=_binary_%s___constants_bin_start"
                % (
                    constant_bin_link_name,
                    "".join(re.sub("[^a-zA-Z0-9_]", "_", c) for c in env.source_dir),
                ),
            ]
        )
    elif resource_mode in ("code", "c23_embed"):
        # Indicate "code" resource mode.
        env.Append(CPPDEFINES=["_NUITKA_CONSTANTS_FROM_CODE"])

        constants_generated_filename = os.path.join(
            env.source_dir, "__constants_data.c"
        )

        def writeConstantsDataSource():
            with openTextFile(constants_generated_filename, "w") as output:
                if not env.c11_mode:
                    output.write('extern "C" {')

                output.write(
                    """\
// Constant data for the program.
"""
                )

                if env.clang_mode or env.clangcl_mode:
                    output.write(
                        """
#if defined(__clang__)
#pragma clang diagnostic ignored "-Wc23-extensions"
#endif
"""
                    )

                output.write(
                    """
#ifdef __cplusplus
extern "C"
#endif
#if !defined(_NUITKA_EXPERIMENTAL_WRITEABLE_CONSTANTS)
const
#endif
unsigned char constant_bin_data[] =\n{\n
"""
                )

                if resource_mode == "code":
                    with open(blob_filename, "rb") as f:
                        content = f.read()
                    for count, stream_byte in enumerate(content):
                        if count % 16 == 0:
                            if count > 0:
                                output.write("\n")

                            output.write("   ")

                        if str is bytes:
                            stream_byte = ord(stream_byte)

                        output.write(" 0x%02x," % stream_byte)
                else:
                    output.write('#embed "%s"\n' % blob_filename)

                output.write("\n};\n")

                if not env.c11_mode:
                    output.write("}")

        writeConstantsDataSource()
    else:
        scons_logger.sysexit(
            "Error, illegal resource mode '%s' specified" % resource_mode
        )


def _enableMacOSTargetSettings(env):
    """Set up environment for macOS target settings."""
    assert isMacOS()

    setEnvironmentVariable(env, "MACOSX_DEPLOYMENT_TARGET", env.macos_min_version)

    # TODO: The zig doesn't support the option given, not having "arm64" as a
    # valid architecture.
    if not env.zig_mode:
        target_flag = "--target=%s-apple-macos%s" % (
            env.macos_target_arch,
            env.macos_min_version,
        )
        env.Append(CCFLAGS=[target_flag])
        env.Append(LINKFLAGS=[target_flag])


def _enableAIXTargetSettings(env):
    """Set up environment for AIX target settings."""
    assert isAIX()

    if env.target_arch in ("64", "32"):
        env.Append(CCFLAGS=["-m%s" % env.target_arch])
        env.Append(LINKFLAGS=["-m%s" % env.target_arch])


def _enableWin32TargetSettings(env):
    """Set up environment for Windows target settings."""
    assert isWin32Windows()
    # The MinGW64 and ClangCL do not default for API level properly, so
    # help it. For zig, it is hard coded.
    if not env.zig_mode:
        env.Append(CPPDEFINES=["_WIN32_WINNT=0x0601"])


def enableWindowsStackSize(env, target_arch):
    # Stack size 4MB or 8MB, we might need more than the default 1MB.
    if target_arch == "x86_64":
        stack_size = 1024 * 1204 * 8
    else:
        stack_size = 1024 * 1204 * 4

    if env.msvc_mode:
        env.Append(LINKFLAGS=["/STACK:%d" % stack_size])

    if env.mingw_mode:
        env.Append(LINKFLAGS=["-Wl,--stack,%d" % stack_size])


def _enableOutputSettings(env):
    if env.msvc_mode:
        # For onefile bootstrap, we have to force static runtime as we don't
        # want to bring a second file and MonolithPy requires it.
        force_static = (
            env.onefile_compile and env.onefile_windows_static_runtime
        ) or env.monolithpy

        if force_static:
            env.Append(CCFLAGS=["/MT"])  # Multithreaded, static version of C run time.
        else:
            env.Append(CCFLAGS=["/MD"])  # Multithreaded, dynamic version of C run time.


def createNuitkaSconsEnvironment(needs_source_dir=True):
    # This is handling the common setup of the Scons environment for Nuitka
    # and has many steps, pylint: disable=too-many-locals,too-many-statements

    # Set up the basic stuff.
    initScons(ARGUMENTS)

    # Helper used for many things.
    job_count = GetOption("num_jobs")

    if needs_source_dir:
        # The directory containing the C files generated by Nuitka to be built using
        # scons. They are referred to as sources from here on.
        source_dir = getArgumentRequired("source_dir")
    else:
        source_dir = "/dev/null"

    # The directory containing Nuitka provided C files to be built and where it
    # should be used.
    nuitka_src = getArgumentRequired("nuitka_src")

    # Debug mode: Less optimizations, debug information in the resulting binary.
    debug_mode = getArgumentBool("debug_mode", False)

    # Debugger mode: Debug information in the resulting binary and intention to run
    # in debugger.
    debugger_mode = getArgumentBool("debugger_mode", False)

    # Experimental indications. Do things that are not yet safe to do.
    experimental_flags = getArgumentList("experimental", "")

    # Deployment mode
    deployment_mode = getArgumentBool("deployment", False)

    # Experimental indications. Do things that are not yet safe to do.
    no_deployment = getArgumentList("no_deployment", "")

    # Debug mode indications. Do check things with fine granularity.
    debug_modes = getArgumentList("debug_modes", "")

    # Tracing mode. Output program progress.
    trace_mode = getArgumentBool("trace_mode", False)

    # LTO mode: Use link time optimizations of C compiler if available and known
    # good with the compiler in question.
    lto_mode = getArgumentDefaulted("lto_mode", "auto")

    # PGO mode: Use profile guided optimization of C compiler if available.
    pgo_mode = getArgumentDefaulted("pgo_mode", "no")

    # Console mode specified
    console_mode = getArgumentDefaulted("console_mode", "attach")

    # Windows might be running a Python whose DLL we have to use.
    uninstalled_python = getArgumentBool("uninstalled_python", False)

    # Unstripped mode: Do not remove debug symbols.
    unstripped_mode = getArgumentBool("unstripped_mode", False)

    # Target arch, uses for compiler choice and quick linking of constants binary
    # data.
    target_arch = getArgumentRequired("target_arch")

    # MinGW compiler mode, optional and interesting to Windows only.
    mingw_mode = getArgumentBool("mingw_mode", False)

    # Clang compiler mode, forced on macOS and FreeBSD (excluding PowerPC), optional on Linux.
    clang_mode = getArgumentBool("clang_mode", False)

    # Clang on Windows with no requirement to use MinGW64 or using MSYS2 MinGW flavor,
    # is changed to ClangCL from Visual Studio.
    clangcl_mode = False
    if os.name == "nt" and not mingw_mode and clang_mode:
        clang_mode = False
        clangcl_mode = True

    # Show scons mode, output information about Scons operation
    show_scons_mode = getArgumentBool("show_scons", False)
    scons_details_logger.is_quiet = not show_scons_mode

    if int(os.getenv("NUITKA_QUIET", "0")):
        from nuitka.Tracing import setQuiet

        setQuiet()

    # Forced MSVC version (windows-only)
    msvc_version = getArgumentDefaulted("msvc_version", None)

    # Allow automatic downloads for ccache, etc.
    assume_yes_for_downloads = getArgumentBool("assume_yes_for_downloads", False)

    # Low memory mode, compile using less memory if possible.
    low_memory = getArgumentBool("low_memory", False)

    # Minimum version required on macOS.
    macos_min_version = getArgumentDefaulted("macos_min_version", "")

    # Target arch for macOS.
    macos_target_arch = getArgumentDefaulted("macos_target_arch", "")

    # gcc compiler cf_protection option
    cf_protection = getArgumentDefaulted("cf_protection", "auto")

    # Amount of frozen modules.
    frozen_modules = getArgumentInt("frozen_modules", 0)

    # Windows runtime DLL inclusion.
    onefile_windows_static_runtime = getArgumentBool(
        "onefile_windows_static_runtime", False
    )

    progress_bar = getArgumentDefaulted("progress_bar", "auto")
    enableSconsProgressBar(progress_bar)

    # Patch the compiler detection.
    Environment.Detect = getEnhancedToolDetect()

    # Create Scons environment, the main control tool. Don't include "mingw" on
    # Windows immediately, we will default to MSVC if available.
    # On Windows, in case MSVC was not found and not previously forced, use the
    # winlibs MinGW64 as a download, and use it as a fallback.
    env = createEnvironmentAndCheckCompiler(
        mingw_mode=mingw_mode,
        msvc_version=msvc_version,
        clang_mode=clang_mode,
        clangcl_mode=clangcl_mode,
        target_arch=target_arch,
        consider_environ_variables=True,
        assume_yes_for_downloads=assume_yes_for_downloads,
        download_ok=True,
        experimental_flags=experimental_flags,
        source_dir=source_dir,
    )

    # Consider switching from gcc to its g++ compiler as a workaround that makes
    # us work without C11.
    switchFromGccToGpp(
        env=env,
    )

    enableFlagSettings(env, "no_deployment", no_deployment)
    env.no_deployment_flags = no_deployment

    enableFlagSettings(env, "experimental", experimental_flags)
    env.experimental_flags = experimental_flags

    enableFlagSettings(env, "debug", debug_modes)
    env.debug_modes_flags = debug_modes

    env.debug_mode = debug_mode
    env.debugger_mode = debugger_mode
    env.unstripped_mode = unstripped_mode
    env.console_mode = console_mode
    env.source_dir = source_dir
    env.deployment_mode = deployment_mode
    env.nuitka_src = nuitka_src
    env.low_memory = low_memory
    env.macos_min_version = macos_min_version
    env.macos_target_arch = macos_target_arch
    env.target_arch = target_arch
    env.experimental = experimental_flags
    env.no_deployment = no_deployment
    env.debug_modes = debug_modes
    env.trace_mode = trace_mode
    env.lto_mode = lto_mode
    env.pgo_mode = pgo_mode
    env.job_count = job_count
    env.frozen_modules = frozen_modules
    env.uninstalled_python = uninstalled_python
    env.onefile_windows_static_runtime = onefile_windows_static_runtime
    env.cf_protection = cf_protection

    if env.the_compiler is None or getExecutablePath(env.the_compiler, env=env) is None:
        raiseNoCompilerFoundErrorExit()

    scons_details_logger.info("Told to run compilation on %d CPUs." % env.job_count)

    if show_scons_mode:
        compiler_path = getExecutablePath(env.the_compiler, env=env)

        if env.msvc_mode:
            scons_logger.info(
                "Scons: Compiler used '%s' (MSVC %s)."
                % (compiler_path, getMsvcVersionString(env))
            )
        else:
            scons_logger.info("Scons: Compiler used '%s'." % compiler_path)

    # Set build directory and scons general settings.
    if needs_source_dir:
        setupScons(env, source_dir)

    return env


def setupCCompiler(env, pgo_mode, exe_target, onefile_compile):
    # This is driven by many branches on purpose and has a lot of things
    # to deal with for LTO checks and flags, pylint: disable=too-many-branches,too-many-statements

    env.exe_target = exe_target
    env.onefile_compile = onefile_compile

    # Enable LTO for compiler.
    _enableLtoSettings(
        env=env,
        lto_mode=env.lto_mode,
        pgo_mode=pgo_mode,
    )

    _enableC11Settings(env)
    _enableOutputSettings(env)

    # Some things are supposed to be only for use with zig, so we inject a
    # define that allows us to test for it.
    if env.zig_mode:
        env.Append(CPPDEFINES=["__ZIG__"])

    if env.gcc_mode or env.zig_mode:
        # Support for gcc and clang, restricting visibility as much as possible.
        env.Append(CCFLAGS=["-fvisibility=hidden"])

        if not env.c11_mode:
            env.Append(CXXFLAGS=["-fvisibility-inlines-hidden"])

        if isWin32Windows() and hasattr(env, "source_dir"):
            # On Windows, exporting to DLL need to be controlled.
            if not env.zig_mode:
                env.Append(LINKFLAGS=["-Wl,--exclude-all-symbols"])

            # Make sure we handle import library on our own and put it into the
            # build directory.
            env.Append(
                LINKFLAGS=[
                    "-Wl,--out-implib,%s" % os.path.join(env.source_dir, "import.lib")
                ]
            )

        # Make it clear how to handle integer overflows, namely by wrapping around
        # to negative values.
        env.Append(CCFLAGS=["-fwrapv"])

        if not env.low_memory:
            # Avoid IO for compilation as much as possible, this should make the
            # compilation more memory hungry, but also faster.
            env.Append(CCFLAGS="-pipe")

        # gcc compiler cf_protection option
        if env.cf_protection != "auto":
            env.Append(CCFLAGS=["-fcf-protection=%s" % env.cf_protection])

    # Support for clang.
    if env.clang_mode or env.clangcl_mode:
        env.Append(CCFLAGS=["-Wno-deprecated-declarations"])

    if isClangName(env.the_cc_name):
        env.Append(CCFLAGS=["-Wno-constant-logical-operand"])

    env.warn_error_mode = (
        env.debug_mode
        and "debug_c_warnings" in env.debug_modes_flags
        and not env.debugger_mode
    )

    if env.warn_error_mode:
        # Allow gcc/clang/MSVC to point out all kinds of inconsistency to us by
        # raising an error.
        if env.gcc_mode or env.zig_mode:
            env.Append(
                CCFLAGS=[
                    "-Wall",
                    "-Werror",  # spell-checker: ignore Werror
                ]
            )

            if env.clang_mode or env.gcc_version >= (4, 6):
                env.Append(CCFLAGS=["-Wunused-but-set-variable"])

        if env.msvc_mode:
            env.Append(CCFLAGS=["/WX"])
    else:
        if env.gcc_mode or env.zig_mode:
            if env.clang_mode or (env.gcc_version and env.gcc_version >= (4, 6)):
                env.Append(CCFLAGS=["-Wno-unused-but-set-variable"])

    # Support for macOS standalone to run on older OS versions.
    if isMacOS():
        _enableMacOSTargetSettings(env)

    if isAIX():
        _enableAIXTargetSettings(env)

        # Otherwise no symbol is exported.
        if env.module_mode or env.dll_mode:
            env.Append(LINKFLAGS=["-Wl,-bexpall"])

        if env.standalone_mode:
            env.Append(LINKFLAGS=["-Wl,-blibpath:/usr/lib:/lib"])

    if isWin32Windows():
        _enableWin32TargetSettings(env)

    # Unicode entry points for programs.
    if env.mingw_mode or (env.zig_mode and isWin32Windows()):
        env.Append(LINKFLAGS=["-municode"])

    # Older g++ complains about aliasing with Py_True and Py_False, but we don't
    # care.
    if (
        (env.gcc_mode or env.zig_mode)
        and not env.clang_mode
        and (env.gcc_version and env.gcc_version < (4, 5))
    ):
        env.Append(CCFLAGS=["-fno-strict-aliasing"])

    # For gcc 4.6 or higher, there this new interesting function.
    if (
        (env.gcc_mode or env.zig_mode)
        and not env.clang_mode
        and (env.gcc_version and env.gcc_version >= (4, 6))
    ):
        env.Append(CCFLAGS=["-fpartial-inlining"])

    # Save some memory for gcc by not tracing macro code locations at all.
    if (
        not env.debug_mode
        and (env.gcc_mode or env.zig_mode)
        and not env.clang_mode
        and env.gcc_version >= (5,)
    ):
        env.Append(CCFLAGS=["-ftrack-macro-expansion=0"])

    # We don't care about deprecations.
    if (env.gcc_mode or env.zig_mode) and not env.clang_mode:
        env.Append(CCFLAGS=["-Wno-deprecated-declarations"])

    # The var-tracking does not scale, disable it. Should we really need it, we
    # can enable it. TODO: Does this cause a performance loss?
    if (env.gcc_mode or env.zig_mode) and not env.clang_mode:
        env.Append(CCFLAGS=["-fno-var-tracking"])

    # For large files, these can issue warnings about disabling
    # itself, while we do not need it really.
    if (
        (env.gcc_mode or env.zig_mode)
        and not env.clang_mode
        and (env.gcc_version and env.gcc_version >= (6,))
    ):
        env.Append(CCFLAGS=["-Wno-misleading-indentation"])

    # Disable output of notes, e.g. on struct alignment layout changes for
    # some arches, we don't care.
    if (env.gcc_mode or env.zig_mode) and not env.clang_mode:
        env.Append(CCFLAGS=["-fcompare-debug-second"])

    # Prevent using LTO when told not to use it, causes errors with some
    # static link libraries.
    if (
        (env.gcc_mode or env.zig_mode)
        and not env.clang_mode
        and env.static_libpython
        and not env.lto_mode
    ):
        env.Append(CCFLAGS=["-fno-lto"])
        env.Append(LINKFLAGS=["-fno-lto"])

    # Set optimization level for gcc and clang in LTO mode
    if (env.gcc_mode or env.zig_mode) and env.lto_mode:
        if env.debug_mode:
            env.Append(LINKFLAGS=["-Og"])
        else:
            # For LTO with static libpython combined, there are crashes with Python core
            # being inlined, so we must refrain from that. On Windows there is no such
            # thing, and MonolithPy is not affected.
            env.Append(
                LINKFLAGS=[
                    (
                        "-O3"
                        if env.monolithpy or os.name == "nt" or not env.static_libpython
                        else "-O2"
                    )
                ]
            )

    # When debugging, optimize less than when optimizing, when not remove
    # assertions.
    if env.debug_mode:
        if env.clang_mode or (
            (env.gcc_mode or env.zig_mode)
            and (env.gcc_version and env.gcc_version >= (4, 8))
        ):
            env.Append(CCFLAGS=["-Og"])
        elif env.gcc_mode or env.zig_mode:
            env.Append(CCFLAGS=["-O1"])
        elif env.msvc_mode:
            env.Append(CCFLAGS=["-O2"])
    else:
        if env.gcc_mode or env.zig_mode:
            env.Append(
                CCFLAGS=[
                    (
                        "-O3"
                        if env.monolithpy or os.name == "nt" or not env.static_libpython
                        else "-O2"
                    )
                ]
            )
        elif env.msvc_mode:
            env.Append(
                CCFLAGS=[
                    "/Ox",  # Enable most speed optimization
                    "/GF",  # Eliminate duplicate strings.
                    "/Gy",  # Function level object storage, to allow removing unused ones
                ]
            )

        env.Append(CPPDEFINES=["__NUITKA_NO_ASSERT__"])

    _enableDebugSystemSettings(env)

    if (env.gcc_mode or env.zig_mode) and not env.noelf_mode:
        env.Append(LINKFLAGS=["-z", "noexecstack"])

    # For MinGW64 we need to tell the subsystem to target as well as to
    # automatically import everything used.
    if env.mingw_mode:
        if not env.clang_mode:
            env.Append(LINKFLAGS=["-Wl,--enable-auto-import"])

    # Even if console is forced, for Win32 it means to specify Windows
    # subsystem, we can still attach or create.
    if env.console_mode in ("attach", "disable"):
        if env.mingw_mode:
            env.Append(LINKFLAGS=["-Wl,--subsystem,windows"])
            env.Append(CPPDEFINES=["_NUITKA_WINMAIN_ENTRY_POINT"])
        elif env.msvc_mode:
            env.Append(LINKFLAGS=["/SUBSYSTEM:windows"])
            env.Append(CPPDEFINES=["_NUITKA_WINMAIN_ENTRY_POINT"])

    if env.console_mode == "attach" and os.name == "nt":
        env.Append(CPPDEFINES=["_NUITKA_ATTACH_CONSOLE_WINDOW"])
        env.Append(LIBS=["User32"])

    if env.console_mode == "disable" and os.name == "nt":
        env.Append(CPPDEFINES=["_NUITKA_DISABLE_CONSOLE_WINDOW"])
        env.Append(LIBS=["User32"])

    if env.console_mode == "hide" and os.name == "nt":
        env.Append(CPPDEFINES=["_NUITKA_HIDE_CONSOLE_WINDOW"])
        env.Append(LIBS=["User32"])

    # Avoid dependency on MinGW libraries, spell-checker: ignore libgcc
    if env.mingw_mode and not env.clang_mode:
        env.Append(LINKFLAGS=["-static-libgcc"])

    # MinGW64 for 64 bits needs this due to CPython bugs.
    if env.mingw_mode and env.target_arch == "x86_64" and env.python_version < (3, 12):
        env.Append(CPPDEFINES=["MS_WIN64"])

    if env.msvc_mode:
        # For shell API usage to lookup app folders we need this.
        env.Append(LIBS=["Shell32"])

        # Make sure to find MSVC "link" in the right spot and to not use the
        # "link" of e.g. Git.
        addToPATH(
            env=env,
            dirname=os.path.dirname(getExecutablePath(env.the_compiler, env=env)),
            prefix=True,
        )

    # Since Fedora 36, the system Python will not link otherwise.
    if isFedoraBasedLinux():
        env.Append(CCFLAGS=["-fPIC"])

    # We use zlib for crc32 functionality
    zlib_inline_copy_dir = os.path.join(env.nuitka_src, "inline_copy", "zlib")
    if os.path.exists(os.path.join(zlib_inline_copy_dir, "crc32.c")):
        env.Append(
            CPPPATH=[
                zlib_inline_copy_dir,
            ],
        )
    else:
        # TODO: Should only happen for official Debian packages, and there we
        # can use the zlib static linking maybe, but for onefile it's not easy
        # to get it, so just use slow checksum for now.
        if onefile_compile:
            env.Append(CPPDEFINES=["_NUITKA_USE_OWN_CRC32"])
        else:
            env.Append(CPPDEFINES=["_NUITKA_USE_SYSTEM_CRC32"])
            env.Append(LIBS="z")

    if isAIX():
        aix_dll_addr_inline_copy_dir = os.path.join(
            env.nuitka_src, "inline_copy", "aix_dll_addr"
        )
        env.Append(
            CPPPATH=[
                aix_dll_addr_inline_copy_dir,
            ],
        )

    # Avoid them as appearing to be different files.
    if (
        env.gcc_mode
        and env.source_dir != "."
        and not getArgumentBool("full_names", False)
    ):
        if env.zig_mode:
            # Zig cc supports it
            file_prefix_map_supported = True
        elif env.clang_mode:
            # Clang 10+ supports --file-prefix-map
            file_prefix_map_supported = env.clang_version >= (10,)
        else:
            # GCC supports it starting from 8.0
            file_prefix_map_supported = env.gcc_version >= (8,)

        if file_prefix_map_supported:
            # TODO: This also lies for modules codes, the mapping should take
            # those into account as well.
            env.Append(
                CCFLAGS=[
                    "--file-prefix-map=%s=%s" % (os.path.normpath(env.source_dir), "."),
                    "--file-prefix-map=%s=%s"
                    % (
                        os.path.normpath(os.path.join(env.source_dir, "static_src")),
                        os.path.normpath(os.path.join(env.nuitka_src, "static_src")),
                    ),
                ]
            )

    if env.msvc_mode:
        # With Clang on Windows, there is also an linker to use.
        env.Append(
            CCFLAGS=[
                "/EHsc",  # No C++ exception handling code.
                "/J",  # default char type is unsigned.
                "/Gd",  # Use C calling convention by default.
                "/bigobj",  # Product object files with larger internal limits, spell-checker: ignore bigobj
            ]
        )

        # No incremental linking.
        env.Append(LINKFLAGS=["/INCREMENTAL:NO"])

        # Remember MSVC installation directory to find redist DLLs, spell-checker: ignore VCINSTALLDIR
        if "VCINSTALLDIR" in env["ENV"]:
            env["MSVC_InstallDirectory"] = getNormalizedPathJoin(
                env["ENV"]["VCINSTALLDIR"], ".."
            )

    # Use compiler/linker flags provided via environment variables, these
    # are always added and consider_environment_variables does not apply
    # since that's about selecting the compiler, not configuring it.
    importEnvironmentVariableSettings(env)


def _enablePgoSettings(env):
    if env.pgo_mode == "no":
        env.progressbar_name = "Backend"
    elif env.pgo_mode == "python":
        env.progressbar_name = "Python Profile"
        env.Append(CPPDEFINES=["_NUITKA_PGO_PYTHON"])
    elif env.pgo_mode == "generate":
        env.progressbar_name = "Profile"
        env.Append(CPPDEFINES=["_NUITKA_PGO_GENERATE"])

        if env.gcc_mode or env.zig_mode:
            env.Append(CCFLAGS=["-fprofile-generate"])
            env.Append(LINKFLAGS=["-fprofile-generate"])
        elif env.msvc_mode:
            env.Append(CCFLAGS=["/GL"])
            env.Append(LINKFLAGS=["/GENPROFILE:EXACT"])
            if not env.clangcl_mode:
                env.Append(LINKFLAGS=["/LTCG"])

        else:
            scons_logger.sysexit(
                "Error, PGO not supported for '%s' compiler." % env.the_cc_name
            )
    elif env.pgo_mode == "use":
        env.progressbar_name = "Backend"

        env.Append(CPPDEFINES=["_NUITKA_PGO_USE"])

        if env.gcc_mode or env.zig_mode:
            env.Append(CCFLAGS=["-fprofile-use"])
            env.Append(LINKFLAGS=["-fprofile-use"])
        elif env.msvc_mode:
            env.Append(CCFLAGS=["/GL"])
            env.Append(
                LINKFLAGS=[
                    "/USEPROFILE",
                ]
            )
        else:
            scons_logger.sysexit(
                "Error, PGO not supported for '%s' compiler." % env.the_cc_name
            )
    else:
        assert False, env.pgo_mode


def _enableDebugSystemSettings(env):
    if env.unstripped_mode:
        # Use debug format, so we get good tracebacks from it.
        if env.gcc_mode or env.zig_mode:
            env.Append(LINKFLAGS=["-g"])
            env.Append(CCFLAGS=["-g"])

            if not env.clang_mode:
                env.Append(CCFLAGS=["-feliminate-unused-debug-types"])
        elif env.msvc_mode:
            env.Append(CCFLAGS=["/Z7"])

            # Higher MSVC versions need this for parallel compilation
            if env.job_count > 1 and getMsvcVersion(env) >= (11,):
                env.Append(CCFLAGS=["/FS"])

            env.Append(LINKFLAGS=["/DEBUG"])
    else:
        if env.gcc_mode or env.zig_mode:
            if isMacOS():
                env.Append(LINKFLAGS=["-Wno-deprecated-declarations"])
            elif not env.clang_mode:
                env.Append(LINKFLAGS=["-s"])


def switchFromGccToGpp(env):
    if not env.gcc_mode or env.clang_mode:
        env.gcc_version = None
        return

    the_compiler = getExecutablePath(env.the_compiler, env)

    if the_compiler is None:
        return

    env.gcc_version = myDetectVersion(env, the_compiler)

    if env.gcc_version is None:
        scons_logger.sysexit(
            """\
Error, failed to detect gcc version of backend compiler '%s'.
"""
            % env.the_compiler
        )

    if "++" in env.the_cc_name:
        scons_logger.sysexit(
            """\
Error, compiler %s is apparently a C++ compiler, specify a C compiler instead.
"""
            % env.the_cc_name
        )

    # Enforce the minimum version, selecting a potentially existing g++-4.5
    # binary if it's not high enough. This is esp. useful under Debian which
    # allows all compiler to exist next to each other and where g++ might not be
    # good enough, but g++-4.5 would be.
    if env.gcc_version < (4, 4):
        scons_logger.sysexit(
            """\
The gcc compiler %s (version %s) doesn't have the sufficient \
version (>= 4.4)."""
            % (env.the_compiler, env.gcc_version)
        )

    # CondaCC or newer.
    if env.mingw_mode and env.gcc_version < (5, 3):
        scons_logger.sysexit(
            """\
The MinGW64 compiler %s (version %s) doesn't have the sufficient \
version (>= 5.3)."""
            % (env.the_compiler, env.gcc_version)
        )

    if env.gcc_version < (5,):
        if env.python_version < (3, 11):
            scons_logger.info(
                "The provided gcc is too old, switching to its g++ instead.",
                mnemonic="too-old-gcc",
            )

            # Switch to g++ from gcc then if possible, when C11 mode is false.
            the_gpp_compiler_name = os.path.basename(env.the_compiler).replace(
                "gcc", "g++"
            )

            the_gpp_compiler = os.path.join(
                os.path.dirname(env.the_compiler),
                the_gpp_compiler_name,
            )

            the_gpp_compiler_path = getExecutablePath(the_gpp_compiler, env=env)

            if the_gpp_compiler_path is None:
                the_gpp_compiler_path = getExecutablePath(
                    the_gpp_compiler_name, env=env
                )

            if the_gpp_compiler_path:
                env.the_compiler = the_gpp_compiler_path
                env.the_cc_name = env.the_cc_name.replace("gcc", "g++")

                env["CC"] = env.the_compiler
                env["CXX"] = env.the_compiler
            else:
                scons_logger.sysexit(
                    """\
Error, your gcc is too old for C11 support, and no g++ ('%s') to
workaround that was found."""
                    % the_gpp_compiler_name
                )
        else:
            scons_logger.sysexit(
                "Error, your gcc is too old for C11 support, install a newer one.",
                mnemonic="too-old-gcc",
            )


def reportCCompiler(env, context, output_func):
    cc_output = env.the_cc_name

    if env.the_cc_name == "cl":
        cc_output = "%s %s" % (env.the_cc_name, getMsvcVersionString(env))
    elif isGccName(env.the_cc_name):
        cc_output = "%s %s" % (
            env.the_cc_name,
            (
                ".".join(str(d) for d in env.gcc_version)
                if env.gcc_version is not None
                else "not found"
            ),
        )
    elif isZigName(env.the_cc_name):
        cc_output = "%s %s" % (
            env.the_cc_name,
            (
                ".".join(str(d) for d in env.zig_version)
                if env.zig_version is not None
                else "not found"
            ),
        )
    elif isClangName(env.the_cc_name):
        cc_output = "%s %s" % (
            env.the_cc_name,
            (
                ".".join(str(d) for d in env.clang_version)
                if env.clang_version is not None
                else "not found"
            ),
        )
    else:
        cc_output = env.the_cc_name

    output_func(
        "%s C compiler: %s (%s)."
        % (context, getReportPath(env.the_compiler), cc_output)
    )


def importEnvironmentVariableSettings(env):
    """Import typical environment variables that compilation should use."""
    # spell-checker: ignore cppflags,cflags,ccflags,cxxflags,ldflags

    env.cpp_flags = os.getenv("CPPFLAGS")
    if env.cpp_flags:
        scons_logger.info("Scons: Inherited CPPFLAGS='%s' variable." % env.cpp_flags)
        env.Append(CPPFLAGS=env.cpp_flags.split())
    env.c_flags = os.getenv("CFLAGS")
    if env.c_flags:
        scons_logger.info("Inherited CFLAGS='%s' variable." % env.c_flags)
        env.Append(CCFLAGS=env.c_flags.split())
    env.cc_flags = os.getenv("CCFLAGS")
    if env.cc_flags:
        scons_logger.info("Inherited CCFLAGS='%s' variable." % env.cc_flags)
        env.Append(CCFLAGS=env.cc_flags.split())
    env.cxx_flags = os.getenv("CXXFLAGS")
    if env.cxx_flags:
        scons_logger.info("Scons: Inherited CXXFLAGS='%s' variable." % env.cxx_flags)
        env.Append(CXXFLAGS=env.cxx_flags.split())

    env.ld_flags = os.getenv("LDFLAGS")
    if env.ld_flags:
        scons_logger.info("Scons: Inherited LDFLAGS='%s' variable." % env.ld_flags)
        env.Append(LINKFLAGS=env.ld_flags.split())


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.gnu.org/licenses/agpl.txt
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2025, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Standard plug-in to make multiprocessing and joblib work well.

On Windows, the multiprocessing modules forks new processes which then have
to start from scratch. This won't work if there is no "sys.executable" to
point to a "Python.exe" and won't use compiled code by default.

The issue applies to accelerated and standalone mode alike.

spell-checker: ignore joblib,anyio
"""

from nuitka.ModuleRegistry import (
    getModuleInclusionInfoByName,
    getRootTopModule,
)
from nuitka.options.Options import isStandaloneMode, shallMakeModule
from nuitka.plugins.PluginBase import NuitkaPluginBase
from nuitka.PythonVersions import python_version
from nuitka.tree.SourceHandling import readSourceCodeFromFilename
from nuitka.utils.ModuleNames import ModuleName


class NuitkaPluginMultiprocessingWorkarounds(NuitkaPluginBase):
    """This is to make multiprocessing work with Nuitka and use compiled code.

    When running in accelerated mode, it's not good to fork a new Python
    instance to run other code, as that won't be accelerated. And when
    run in standalone mode, there may not even be a Python, but it's the
    same principle.

    So by default, this module is on and works around the behavior of the
    "multiprocessing.forking/multiprocessing.spawn/multiprocessing.manager"
    expectations.
    """

    plugin_name = "multiprocessing"
    plugin_desc = "Required by 'multiprocessing' package."
    plugin_category = "package-support"

    @classmethod
    def isRelevant(cls):
        return not shallMakeModule()

    @staticmethod
    def isAlwaysEnabled():
        return True

    @staticmethod
    def createPreModuleLoadCode(module):
        full_name = module.getFullName()

        if full_name == "__parents_main__":
            yield (
                """\
import sys
# If the multiprocessing fork flag is present, scrub the arguments so that the
# user code (which might use argparse) doesn't see them. Save them in builtins
# so we can restore them later when calling multiprocessing.spawn.spawn_main.
if "--multiprocessing-fork" in sys.argv:
    # Use builtins to store the args so they are available globally even in functions
    # without needing to pass them around or use global keyword everywhere.
    if str is bytes:
        import __builtin__ as builtins
    else:
        import builtins

    builtins.__nuitka_original_args = sys.argv[:]
    sys.argv = [sys.argv[0]]
""",
                "Scrubbing multiprocessing fork arguments from sys.argv.",
            )

        if full_name in ("multiprocessing", "anyio"):
            yield (
                """\
import sys, os
argv0 = sys.argv[0]
if sys.platform == "win32" and not os.path.exists(argv0) and not argv0.endswith(".exe"):
    argv0 += ".exe"

sys.executable = %s
sys._base_executable = sys.executable
"""
                % ("__nuitka_binary_exe" if isStandaloneMode() else "argv0"),
                """\
Monkey patching "%s" load environment."""
                % full_name,
            )

    @staticmethod
    def createPostModuleLoadCode(module):
        full_name = module.getFullName()

        if full_name == "multiprocessing":
            yield (
                """\
try:
    from multiprocessing.forking import ForkingPickler
except ImportError:
    from multiprocessing.reduction import ForkingPickler

class C:
   def f():
       pass

def _reduce_compiled_method(m):
    if m.im_self is None:
        return getattr, (m.im_class, m.im_func.__name__)
    else:
        return getattr, (m.im_self, m.im_func.__name__)

ForkingPickler.register(type(C().f), _reduce_compiled_method)
if str is bytes:
    ForkingPickler.register(type(C.f), _reduce_compiled_method)
""",
                """\
Monkey patching "multiprocessing" for compiled methods.""",
            )

    @staticmethod
    def createFakeModuleDependency(module):
        full_name = module.getFullName()

        if full_name != "multiprocessing":
            return

        # First, build the module node and then read again from the
        # source code.
        root_module = getRootTopModule()

        # TODO: Might have to also make "__mp_main__"
        module_name = ModuleName("__parents_main__")

        source_code = readSourceCodeFromFilename(module_name, root_module.getFilename())

        # For the call stack, this may look bad or different to what CPython
        # does, but such is life. The forking module was split up into multiple
        # modules in Python3 so this is doubled. TODO: Move the version check
        # into the code and let Nuitka compile time optimize it, which will also
        # make sure joblib support is there for Python2 as well.
        if python_version >= 0x300:
            source_code += """
def __nuitka_freeze_support():
    import sys
    import builtins

    # Not needed, and can crash from minor "__file__" differences, depending on invocation
    import multiprocessing.spawn
    multiprocessing.spawn._fixup_main_from_path = lambda mod_name : None

    # This is a variant of freeze_support that will work for multiprocessing and
    # joblib equally well.
    kwds = {}
    args = []

    if hasattr(builtins, "__nuitka_original_args"):
        sys.argv = builtins.__nuitka_original_args
        del builtins.__nuitka_original_args

    for arg in sys.argv[2:]:
        try:
            name, value = arg.split('=')
        except ValueError:
            name = "pipe_handle"
            value = arg

        if value == 'None':
            kwds[name] = None
        else:
            kwds[name] = int(value)

    # Otherwise main module names will not work.
    sys.modules["__main__"] = sys.modules["__parents_main__"]

    multiprocessing.spawn.spawn_main(*args, **kwds)
__nuitka_freeze_support()
"""
        else:
            source_code += """
def __nuitka_freeze_support():
    import __builtin__, sys
    if hasattr(__builtin__, "__nuitka_original_args"):
        sys.argv = __builtin__.__nuitka_original_args
        del __builtin__.__nuitka_original_args

    sys.modules["__main__"] = sys.modules[__name__]
    __import__("multiprocessing.forking").forking.freeze_support()
__nuitka_freeze_support()
"""

        yield (
            module_name,
            source_code,
            root_module.getCompileTimeFilename(),
            "Auto enable multiprocessing freeze support",
        )

    def onModuleEncounter(
        self, using_module_name, module_name, module_filename, module_kind
    ):
        # Enforce recursion in to multiprocessing for accelerated mode, which
        # would normally avoid this.
        if module_name.hasNamespace("multiprocessing"):
            return True, "Multiprocessing plugin needs this to monkey patch it."

    def decideCompilation(self, module_name):
        if module_name.hasNamespace("multiprocessing"):
            return "bytecode"

        # TODO: Make this demoted too.
        # or module_name in( "multiprocessing-preLoad", "multiprocessing-postLoad"):

    @staticmethod
    def getPreprocessorSymbols():
        if getModuleInclusionInfoByName("__parents_main__"):
            return {"_NUITKA_PLUGIN_MULTIPROCESSING_ENABLED": "1"}


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.gnu.org/licenses/agpl.txt
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2025, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Standard plug-in to find data files."""

import os
import pkgutil

from nuitka.code_generation.ConstantCodes import addDistributionMetadataValue
from nuitka.containers.OrderedSets import OrderedSet
from nuitka.options.Options import isStandaloneMode
from nuitka.plugins.YamlPluginBase import NuitkaYamlPluginBase
from nuitka.PythonFlavors import isDebianPackagePython
from nuitka.utils.Distributions import getDistribution
from nuitka.utils.FileOperations import (
    changeFilenameExtension,
    getFileList,
    resolveShellPatternToFilenames,
)


class NuitkaPluginDataFileCollector(NuitkaYamlPluginBase):
    plugin_name = "data-files"
    plugin_desc = "Include data files specified by package configuration files."
    plugin_category = "core,feature"

    @classmethod
    def isRelevant(cls):
        return isStandaloneMode()

    @staticmethod
    def isAlwaysEnabled():
        return True

    def _considerDataFiles(self, module, data_file_config):
        # Many details and cases to deal with
        # pylint: disable=too-many-branches,too-many-locals,too-many-statements

        module_name = module.getFullName()
        module_folder = module.getCompileTimeDirectory()

        target_dir = data_file_config.get("dest_path")

        # Default to near module or inside package folder.
        if target_dir is None:
            if (
                module.isCompiledPythonPackage()
                or module.isUncompiledPythonPackage()
                or module.isExtensionModulePackage()
            ):
                target_dir = module_name.asPath()
            else:
                package_name = module_name.getPackageName()

                if package_name is not None:
                    target_dir = module_name.getPackageName().asPath()
                else:
                    target_dir = "."

        patterns = data_file_config.get("patterns")
        if patterns is not None:
            if type(patterns) is not list or not patterns:
                self.sysexit(
                    "Error, requiring list below 'pattern' entry for '%s' entry."
                    % module_name
                )

            # TODO: Pattern should be data file kind potentially.
            for pattern in patterns:
                pattern = self.evaluateExpressionOrConstant(
                    full_name=module_name,
                    expression=pattern,
                    config_name="data file pattern for '%s'" % module_name,
                    extra_context=None,
                    single_value=True,
                )

                pattern = os.path.join(module_folder, pattern)

                for filename in resolveShellPatternToFilenames(pattern):
                    filename_base = os.path.relpath(filename, module_folder)

                    yield self.makeIncludedDataFile(
                        source_path=filename,
                        dest_path=os.path.normpath(
                            os.path.join(target_dir, filename_base)
                        ),
                        reason="package data for '%s'" % module_name.asString(),
                        tags="config",
                    )

        empty_dirs = data_file_config.get("empty_dirs")
        if empty_dirs is not None:
            if type(empty_dirs) is not list or not empty_dirs:
                self.sysexit(
                    "Error, requiring list below 'empty_dirs' entry for '%s' entry."
                    % module_name
                )

            for empty_dir in empty_dirs:
                yield self.makeIncludedEmptyDirectory(
                    dest_path=os.path.normpath(os.path.join(target_dir, empty_dir)),
                    reason="empty dir needed for %r" % module_name.asString(),
                    tags="config",
                )

        empty_dir_structures = data_file_config.get("empty_dir_structures")
        if empty_dir_structures is not None:
            if type(empty_dir_structures) is not list or not empty_dir_structures:
                self.sysexit(
                    "Error, requiring list below 'empty_dirs_structure' entry for '%s' entry."
                    % module_name
                )

            # TODO: This ignored config dest_path, which is unused, but not consistent.
            for included_data_file in self._getSubDirectoryFolders(
                module, sub_dirs=empty_dir_structures
            ):
                yield included_data_file

        dirs = data_file_config.get("dirs")
        if dirs is not None:
            if type(dirs) is not list or not dirs:
                self.sysexit(
                    "Error, requiring list below 'dirs' entry for '%s' entry."
                    % module_name
                )

            for data_dir in dirs:
                source_path = os.path.join(module_folder, data_dir)

                if os.path.isdir(source_path):
                    yield self.makeIncludedDataDirectory(
                        source_path=source_path,
                        dest_path=os.path.normpath(os.path.join(target_dir, data_dir)),
                        reason="package data directory '%s' for %r"
                        % (data_dir, module_name.asString()),
                        tags="config",
                    )

        raw_dirs = data_file_config.get("raw_dirs")
        if raw_dirs is not None:
            if type(raw_dirs) is not list or not raw_dirs:
                self.sysexit(
                    "Error, requiring list below 'raw_dirs' entry for '%s' entry."
                    % module_name
                )

            for raw_dir in raw_dirs:
                source_path = os.path.join(module_folder, raw_dir)

                if os.path.isdir(source_path):
                    yield self.makeIncludedDataDirectory(
                        source_path=source_path,
                        dest_path=os.path.normpath(os.path.join(target_dir, raw_dir)),
                        reason="package raw directory '%s' for %r"
                        % (raw_dir, module_name.asString()),
                        tags="config",
                        raw=True,
                    )

        include_pyi_file = data_file_config.get("include-pyi-file")

        if include_pyi_file == "yes":
            pyi_filename = changeFilenameExtension(
                path=module.getCompileTimeFilename(), extension=".pyi"
            )

            if os.path.exists(pyi_filename):
                if (
                    module.isCompiledPythonPackage()
                    or module.isUncompiledPythonPackage()
                ):
                    module_path = module_name.asPath()
                else:
                    module_path = os.path.dirname(module_name.asPath())

                yield self.makeIncludedDataFile(
                    source_path=pyi_filename,
                    dest_path=os.path.join(module_path, os.path.basename(pyi_filename)),
                    reason="runtime required '.pyi' file for '%s'"
                    % module_name.asString(),
                    tags="config",
                )

        distribution_names = data_file_config.get("include-metadata", ())

        for distribution_name in distribution_names:
            distribution = getDistribution(distribution_name)

            if distribution is not None:
                addDistributionMetadataValue(
                    distribution_name=distribution_name,
                    distribution=distribution,
                    reason="According to Yaml configuration 'include-metadata' for '%s'"
                    % module_name.asString(),
                )

    def considerDataFiles(self, module):
        full_name = module.getFullName()

        for entry in self.config.get(full_name, section="data-files"):
            if self.evaluateCondition(
                full_name=full_name, condition=entry.get("when", "True")
            ):
                for included_data_file in self._considerDataFiles(
                    module=module, data_file_config=entry
                ):
                    yield included_data_file

        # TODO: Until the data files are a list and support features to do similar, namely
        # to look up via package data interface "pkgutil.get_data" rather than file scan.
        if full_name == "lib2to3.pygram" and isDebianPackagePython():
            yield self.makeIncludedGeneratedDataFile(
                data=pkgutil.get_data("lib2to3", "Grammar.txt"),
                dest_path="lib2to3/Grammar.txt",
                reason="package data for '%s'" % full_name,
                tags="config",
            )

            yield self.makeIncludedGeneratedDataFile(
                data=pkgutil.get_data("lib2to3", "PatternGrammar.txt"),
                dest_path="lib2to3/PatternGrammar.txt",
                reason="package data for '%s'" % full_name,
                tags="config",
            )

    def _getSubDirectoryFolders(self, module, sub_dirs):
        """Get dirnames in given subdirectories of the module.

        Notes:
            All dirnames in folders below one of the sub_dirs are recursively
            retrieved and returned shortened to begin with the string of subdir.
        Args:
            module: module object
            sub_dirs: sub folder name(s) - tuple
        Returns:
            makeIncludedEmptyDirectory of found dirnames.
        """

        module_dir = module.getCompileTimeDirectory()
        file_list = []

        data_dirs = [os.path.join(module_dir, subdir) for subdir in sub_dirs]

        # Gather the full file list, probably makes no sense to include bytecode files
        file_list = sum(
            (
                getFileList(
                    data_dir, ignore_dirs=("__pycache__",), ignore_suffixes=(".pyc",)
                )
                for data_dir in data_dirs
            ),
            [],
        )

        if not file_list:
            msg = "No files or folders found for '%s' in subfolder(s) '%s' (%r)." % (
                module.getFullName(),
                sub_dirs,
                data_dirs,
            )
            self.warning(msg)

        is_package = (
            module.isCompiledPythonPackage() or module.isUncompiledPythonPackage()
        )

        # We need to preserve the package target path in the dist folder.
        if is_package:
            package_part = module.getFullName().asPath()
        else:
            package = module.getFullName().getPackageName()

            if package is None:
                package_part = ""
            else:
                package_part = package.asPath()

        item_set = OrderedSet()

        for f in file_list:
            target = os.path.join(package_part, os.path.relpath(f, module_dir))

            dir_name = os.path.dirname(target)
            item_set.add(dir_name)

        for dest_path in item_set:
            yield self.makeIncludedEmptyDirectory(
                dest_path=dest_path,
                reason="Subdirectories of module %s" % module.getFullName(),
                tags="config",
            )


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.gnu.org/licenses/agpl.txt
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2025, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Display the Distributions installed."""

import os

from nuitka.containers.OrderedSets import OrderedSet
from nuitka.importing.Importing import (
    addMainScriptDirectory,
    hasMainScriptDirectory,
)
from nuitka.Tracing import my_print
from nuitka.utils.Distributions import (
    getDistributionInstallerName,
    getDistributionName,
    getDistributions,
    getDistributionVersion,
)


def displayDistributions():
    if not hasMainScriptDirectory():
        addMainScriptDirectory(os.getcwd())

    output = OrderedSet()

    for distributions in getDistributions().values():
        for distribution in distributions:
            distribution_name = getDistributionName(distribution)

            output.add(
                (
                    distribution_name,
                    getDistributionVersion(distribution),
                    getDistributionInstallerName(distribution_name=distribution_name),
                )
            )

    for item in sorted(output):
        my_print(*item)


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.gnu.org/licenses/agpl.txt
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2025, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Reformulation of "yield" and "yield from" expressions.

Consult the Developer Manual for information. TODO: Add ability to sync
source code comments with Developer Manual sections.

"""

import ast

from nuitka.nodes.ConstantRefNodes import ExpressionConstantNoneRef
from nuitka.nodes.CoroutineNodes import ExpressionAsyncWait
from nuitka.nodes.YieldNodes import (
    ExpressionYield,
    ExpressionYieldFrom,
    ExpressionYieldFromAwaitable,
)
from nuitka.PythonVersions import python_version

from .SyntaxErrors import raiseSyntaxError
from .TreeHelpers import buildNode


def _getErrorMessageYieldFromOutsideFunction():
    # Need to use "exec" to detect the syntax error, pylint: disable=W0122

    try:
        exec("""yield from ()""")
    except SyntaxError as e:
        if "yield from" in str(e):
            return "yield from"
        else:
            return "yield"


def _checkInsideGenerator(construct_name, provider, node, source_ref):

    if provider.isCompiledPythonModule():
        # Bug compatibility
        if construct_name == "yield from":
            construct_error_name = _getErrorMessageYieldFromOutsideFunction()
        else:
            construct_error_name = construct_name

        raiseSyntaxError(
            "'%s' outside function" % construct_error_name,
            source_ref.atColumnNumber(node.col_offset),
        )

    # This yield is forbidden in 3.5, but allowed in 3.6, but "yield from"
    # is neither.
    if provider.isExpressionAsyncgenObjectBody() and construct_name == "yield from":
        raiseSyntaxError(
            "'%s' inside async function"
            % ("yield" if node.__class__ is ast.Yield else "yield from",),
            source_ref.atColumnNumber(node.col_offset),
        )

    if (
        python_version >= 0x380
        and provider.isExpressionGeneratorObjectBody()
        and provider.name == "<genexpr>"
        and construct_name != "await"
    ):
        raiseSyntaxError(
            "'%s' inside generator expression"
            % ("yield" if node.__class__ is ast.Yield else "yield from",),
            provider.getSourceReference(),
        )

    while provider.isExpressionOutlineFunction():
        provider = provider.getParentVariableProvider()

    assert (
        provider.isExpressionGeneratorObjectBody()
        or provider.isExpressionAsyncgenObjectBody()
        or provider.isExpressionCoroutineObjectBody()
    ), provider


def buildYieldNode(provider, node, source_ref):
    _checkInsideGenerator("yield", provider, node, source_ref)

    if node.value is not None:
        return ExpressionYield(
            expression=buildNode(provider, node.value, source_ref),
            source_ref=source_ref,
        )
    else:
        return ExpressionYield(
            expression=ExpressionConstantNoneRef(source_ref=source_ref),
            source_ref=source_ref,
        )


def buildYieldFromNode(provider, node, source_ref):
    assert python_version >= 0x300

    _checkInsideGenerator("yield from", provider, node, source_ref)

    return ExpressionYieldFrom(
        expression=buildNode(provider, node.value, source_ref), source_ref=source_ref
    )


def buildAwaitNode(provider, node, source_ref):
    _checkInsideGenerator("await", provider, node, source_ref)

    return ExpressionYieldFromAwaitable(
        expression=ExpressionAsyncWait(
            expression=buildNode(provider, node.value, source_ref),
            source_ref=source_ref,
        ),
        source_ref=source_ref,
    )


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.gnu.org/licenses/agpl.txt
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



# SPDX-License-Identifier: MIT
#
# Copyright The SCons Foundation

"""Various SCons utility functions

Routines which check types and do type conversions.
"""
from __future__ import annotations

import codecs
import os
import pprint
import re
import sys

from collections import UserDict, UserList, UserString, deque
from collections.abc import Iterable, KeysView, ValuesView, ItemsView

# Functions for deciding if things are like various types, mainly to
# handle UserDict, UserList and UserString like their underlying types.
#
# Yes, all of this manual testing breaks polymorphism, and the real
# Pythonic way to do all of this would be to just try it and handle the
# exception, but handling the exception when it's not the right type is
# often too slow.

# A trick is used to speed up these functions. Default arguments are
# used to take a snapshot of the global functions and constants used
# by these functions. This transforms accesses to global variables into
# local variable accesses (i.e. LOAD_FAST instead of LOAD_GLOBAL).
# Since checkers dislike this, it's now annotated for pylint, to flag
# (mostly for other readers of this code) we're doing this intentionally.
# TODO: experts affirm this is still faster, but maybe check if worth it?

DictTypes = (dict, UserDict)
ListTypes = (list, UserList, deque)

# With Python 3, there are view types that are sequences. Other interesting
# sequences are range and bytearray.  What we don't want is strings: while
# they are iterable sequences, in SCons usage iterating over a string is
# almost never what we want. So basically iterable-but-not-string:
SequenceTypes = (list, tuple, deque, UserList, KeysView, ValuesView, ItemsView)

# Note that profiling data shows a speed-up when comparing
# explicitly with str instead of simply comparing
# with basestring. (at least on Python 2.5.1)
# TODO: PY3 check this benchmarking is still correct.
StringTypes = (str, UserString)

# Empirically, it is faster to check explicitly for str than for basestring.
BaseStringTypes = str

# Later Python versions allow us to explicitly apply type hints based off the
# return value similar to isinstance(), albeit not as precise.
if sys.version_info >= (3, 13):
    from typing import TypeAlias, TypeIs

    DictTypeRet: TypeAlias = TypeIs[dict | UserDict]
    ListTypeRet: TypeAlias = TypeIs[list | UserList | deque]
    SequenceTypeRet: TypeAlias = TypeIs[list | tuple | deque | UserList | KeysView | ValuesView | ItemsView]
    TupleTypeRet: TypeAlias = TypeIs[tuple]
    StringTypeRet: TypeAlias = TypeIs[str | UserString]
elif sys.version_info >= (3, 10):
    from typing import TypeAlias, TypeGuard

    DictTypeRet: TypeAlias = TypeGuard[dict | UserDict]
    ListTypeRet: TypeAlias = TypeGuard[list | UserList | deque]
    SequenceTypeRet: TypeAlias = TypeGuard[list | tuple | deque | UserList | KeysView | ValuesView | ItemsView]
    TupleTypeRet: TypeAlias = TypeGuard[tuple]
    StringTypeRet: TypeAlias = TypeGuard[str | UserString]
else:
    # Because we have neither `TypeAlias` class nor `type` keyword pre-3.10,
    # the boolean fallback type has to be wrapped in the legacy `Union` class.
    from typing import Union

    DictTypeRet = Union[bool, bool]
    ListTypeRet = Union[bool, bool]
    SequenceTypeRet = Union[bool, bool]
    TupleTypeRet = Union[bool, bool]
    StringTypeRet = Union[bool, bool]


def is_Dict(  # pylint: disable=redefined-outer-name,redefined-builtin
    obj, isinstance=isinstance, DictTypes=DictTypes
):
    """Check if object is a dict."""
    return isinstance(obj, DictTypes)


def is_List(  # pylint: disable=redefined-outer-name,redefined-builtin
    obj, isinstance=isinstance, ListTypes=ListTypes
):
    """Check if object is a list."""
    return isinstance(obj, ListTypes)


def is_Sequence(  # pylint: disable=redefined-outer-name,redefined-builtin
    obj, isinstance=isinstance, SequenceTypes=SequenceTypes
):
    """Check if object is a sequence."""
    return isinstance(obj, SequenceTypes)


def is_Tuple(  # pylint: disable=redefined-builtin
    obj, isinstance=isinstance, tuple=tuple
):
    """Check if object is a tuple."""
    return isinstance(obj, tuple)


def is_String(  # pylint: disable=redefined-outer-name,redefined-builtin
    obj, isinstance=isinstance, StringTypes=StringTypes
):
    """Check if object is a string."""
    return isinstance(obj, StringTypes)


def is_Scalar(  # pylint: disable=redefined-outer-name,redefined-builtin
    obj, isinstance=isinstance, StringTypes=StringTypes, Iterable=Iterable,
):
    """Check if object is a scalar: not a container or iterable."""
    # Profiling shows that there is an impressive speed-up of 2x
    # when explicitly checking for strings instead of just not
    # sequence when the argument (i.e. obj) is already a string.
    # But, if obj is a not string then it is twice as fast to
    # check only for 'not sequence'. The following code therefore
    # assumes that the obj argument is a string most of the time.
    # Update: now using collections.abc.Iterable for the 2nd check.
    # Note: None is considered a "scalar" for this check, which is correct
    # for the usage in SCons.Environment._add_cppdefines.
    return isinstance(obj, StringTypes) or not isinstance(obj, Iterable)


# From Dinu C. Gherman,
# Python Cookbook, second edition, recipe 6.17, p. 277.
# Also: https://code.activestate.com/recipes/68205
# ASPN: Python Cookbook: Null Object Design Pattern


class Null:
    """Null objects always and reliably 'do nothing'."""

    def __new__(cls, *args, **kwargs):
        if '_instance' not in vars(cls):
            cls._instance = super().__new__(cls, *args, **kwargs)
        return cls._instance

    def __init__(self, *args, **kwargs):
        pass

    def __call__(self, *args, **kwargs):
        return self

    def __repr__(self):
        return f"Null(0x{id(self):08X})"

    def __bool__(self):
        return False

    def __getattr__(self, name):
        return self

    def __setattr__(self, name, value):
        return self

    def __delattr__(self, name):
        return self


class NullSeq(Null):
    """A Null object that can also be iterated over."""

    def __len__(self):
        return 0

    def __iter__(self):
        return iter(())

    def __getitem__(self, i):
        return self

    def __delitem__(self, i):
        return self

    def __setitem__(self, i, v):
        return self


def to_bytes(s):
    """Convert object to bytes."""
    if s is None:
        return b'None'
    if isinstance(s, (bytes, bytearray)):
        # if already bytes return.
        return s
    return bytes(s, 'utf-8')


def to_str(s):
    """Convert object to string."""
    if s is None:
        return 'None'
    if is_String(s):
        return s
    return str(s, 'utf-8')


# Generic convert-to-string functions.  The wrapper
# to_String_for_signature() will use a for_signature() method if the
# specified object has one.


def to_String(  # pylint: disable=redefined-outer-name,redefined-builtin
    obj,
    isinstance=isinstance,
    str=str,
    UserString=UserString,
    BaseStringTypes=BaseStringTypes,
):
    """Return a string version of obj.

    Use this for data likely to be well-behaved. Use
    :func:`to_Text` for unknown file data that needs to be decoded.
    """
    if isinstance(obj, BaseStringTypes):
        # Early out when already a string!
        return obj

    if isinstance(obj, UserString):
        # obj.data can only be a regular string. Please see the UserString initializer.
        return obj.data

    return str(obj)


def to_String_for_subst(  # pylint: disable=redefined-outer-name,redefined-builtin
    obj,
    isinstance=isinstance,
    str=str,
    BaseStringTypes=BaseStringTypes,
    SequenceTypes=SequenceTypes,
    UserString=UserString,
):
    """Return a string version of obj for subst usage."""
    # Note that the test cases are sorted by order of probability.
    if isinstance(obj, BaseStringTypes):
        return obj

    if isinstance(obj, SequenceTypes):
        return ' '.join([to_String_for_subst(e) for e in obj])

    if isinstance(obj, UserString):
        # obj.data can only a regular string. Please see the UserString initializer.
        return obj.data

    return str(obj)


def to_String_for_signature(  # pylint: disable=redefined-outer-name,redefined-builtin
    obj, to_String_for_subst=to_String_for_subst, AttributeError=AttributeError,
):
    """Return a string version of obj for signature usage.

    Like :func:`to_String_for_subst` but has special handling for
    scons objects that have a :meth:`for_signature` method, and for dicts.
    """
    try:
        f = obj.for_signature
    except AttributeError:
        if isinstance(obj, dict):
            # pprint will output dictionary in key sorted order
            # with py3.5 the order was randomized. Depending on dict order
            # which was undefined until py3.6 (where it's by insertion order)
            # was not wise.
            # TODO: Change code when floor is raised to PY36
            return pprint.pformat(obj, width=1000000)
        return to_String_for_subst(obj)
    return f()


def to_Text(data):
    """Return bytes data converted to text.

    Useful for whole-file reads where the data needs some interpretation,
    particularly for Scanners.  Attempts to figure out what the encoding of
    the text is based upon the BOM bytes, and then decodes the contents so
    that it's a valid python string.
    """
    _encoding_map = [
        (codecs.BOM_UTF8, 'utf-8'),
        (codecs.BOM_UTF16_LE, 'utf-16le'),
        (codecs.BOM_UTF16_BE, 'utf-16be'),
        (codecs.BOM_UTF32_LE, 'utf-32le'),
        (codecs.BOM_UTF32_BE, 'utf-32be'),
    ]

    # First look for Byte-order-mark sequences to identify the encoding.
    # Strip these since some codecs do, some don't.
    for bom, encoding in _encoding_map:
        if data.startswith(bom):
            return data[len(bom):].decode(encoding, errors='backslashreplace')

    # If we didn't see a BOM, try UTF-8, then the "preferred" encoding
    # (the files might be written on this system), then finally latin-1.
    # TODO: possibly should be a way for the build to set an encoding.
    try:
        return data.decode('utf-8')
    except UnicodeDecodeError:
            try:
                import locale
                prefencoding = locale.getpreferredencoding()
                return data.decode(prefencoding)
            except (UnicodeDecodeError, LookupError):
                return data.decode('latin-1', errors='backslashreplace')


def get_env_bool(env, name, default=False):
    """Convert a construction variable to bool.

    If the value of *name* in dict-like object *env* is 'true', 'yes',
    'y', 'on' (case insensitive) or anything convertible to int that
    yields non-zero, return ``True``; if 'false', 'no', 'n', 'off'
    (case insensitive) or a number that converts to integer zero return
    ``False``.  Otherwise, or if *name* is not found, return the value
    of *default*.

    Args:
        env: construction environment, or any dict-like object.
        name: name of the variable.
        default: value to return if *name* not in *env* or cannot
          be converted (default: False).
    """
    try:
        var = env[name]
    except KeyError:
        return default
    try:
        return bool(int(var))
    except ValueError:
        if str(var).lower() in ('true', 'yes', 'y', 'on'):
            return True

        if str(var).lower() in ('false', 'no', 'n', 'off'):
            return False

        return default


def get_os_env_bool(name, default=False):
    """Convert an external environment variable to boolean.

    Like :func:`get_env_bool`, but uses :attr:`os.environ` as the lookup dict.
    """
    return get_env_bool(os.environ, name, default)


_get_env_var = re.compile(r'^\$([_a-zA-Z]\w*|{[_a-zA-Z]\w*})$')


def get_environment_var(varstr):
    """Return undecorated construction variable string.

    Determine if *varstr* looks like a reference
    to a single environment variable, like ``"$FOO"`` or ``"${FOO}"``.
    If so, return that variable with no decorations, like ``"FOO"``.
    If not, return ``None``.
    """
    mo = _get_env_var.match(to_String(varstr))
    if mo:
        var = mo.group(1)
        if var[0] == '{':
            return var[1:-1]
        return var

    return None


# Local Variables:
# tab-width:4
# indent-tabs-mode:nil
# End:
# vim: set expandtab tabstop=4 shiftwidth=4:



#     Copyright 2025, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""This test uses the meta data of metadata entry points lookups that wouldn't work at runtime.

But since Nuitka resolves them at compile time, no issue should happen.
"""

# nuitka-project: --mode=standalone

import warnings

warnings.filterwarnings("ignore", category=DeprecationWarning)

try:
    from importlib.metadata import entry_points
except ImportError:
    pass
else:
    print("Type of importlib.metadata.entry_points", type(entry_points()))
    try:
        print(
            "Type of importlib.metadata.entry_points subscript",
            type(entry_points()["console_scripts"]),
        )
        print("There are %d entry points" % len(entry_points()["console_scripts"]))
        print(
            "Example entry point",
            sorted(entry_points()["console_scripts"], key=lambda e: e.name)[0],
        )
    except KeyError:
        pass

try:
    from importlib_metadata import entry_points
except ImportError:
    pass
else:
    print("Type of importlib_metadata.entry_points", type(entry_points()))
    try:
        print(
            "Type of importlib_metadata.entry_points subscript",
            type(entry_points()["console_scripts"]),
        )
        print("There are %d entry points" % len(entry_points()["console_scripts"]))
        print(
            "Example entry point",
            sorted(entry_points()["console_scripts"], key=lambda e: e.name)[0],
        )
    except KeyError:
        pass
    try:
        print(
            "Example group from importlib_metadata.entry_points element",
            sorted(entry_points().groups)[0],
        )
    except (IndexError, AttributeError):
        pass

#     Python test originally created or extracted from other peoples work. The
#     parts from me are licensed as below. It is at least Free Software where
#     it's copied from other people. In these cases, that will normally be
#     indicated.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2025, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Helper functions for the scons file."""

from __future__ import print_function

import os
import pickle
import signal
import sys

from nuitka.__past__ import (  # pylint: disable=redefined-builtin
    FileNotFoundError,
    basestring,
    unicode,
)
from nuitka.containers.OrderedDicts import OrderedDict
from nuitka.Progress import enableThreading
from nuitka.PythonFlavors import isTermuxPython
from nuitka.Tracing import scons_details_logger, scons_logger
from nuitka.utils.Execution import executeProcess
from nuitka.utils.FileOperations import (
    changeFilenameExtension,
    deleteFile,
    getExternalUsePath,
    getFileContentByLine,
    getFilenameExtension,
    getNormalizedPath,
    getNormalizedPathJoin,
    getWindowsShortPathName,
    hasFilenameExtension,
    isFilesystemEncodable,
    openPickleFile,
    openTextFile,
    withFileLock,
)
from nuitka.utils.Utils import isLinux, isMacOS, isPosixWindows, isWin32Windows


def initScons(arguments):
    # Set the arguments.
    _setArguments(arguments)

    # Avoid localized outputs, spell-checker: ignore VSLANG
    os.environ["LC_ALL"] = "C"
    os.environ["VSLANG"] = "1033"

    def no_sync(self):
        # That's a noop, pylint: disable=unused-argument
        pass

    # Avoid scons writing the scons database at all, we don't care for it,
    # spell-checker: ignore dblite
    import SCons.dblite  # pylint: disable=I0021,import-error

    try:
        SCons.dblite.dblite.sync = no_sync
    except AttributeError:
        SCons.dblite._Dblite.sync = no_sync  # pylint: disable=protected-access

    # We use threads during build, so keep locks if necessary for progress bar
    # updates.
    enableThreading()


scons_arguments = {}


def _setArguments(arguments):
    """Decode command line arguments."""

    arg_encoding = arguments.get("argument_encoding")

    for key, value in arguments.items():
        if arg_encoding is not None:
            value = decodeData(value)
        scons_arguments[key] = value


def setupScons(env, source_dir):
    env["BUILD_DIR"] = source_dir

    # Store the file signatures database with the rest of the source files and
    # make it version dependent on the Python version of running Scons, as its
    # pickle is being used, spell-checker: ignore sconsign
    sconsign_filename = os.path.abspath(
        getNormalizedPathJoin(
            source_dir, ".sconsign-%d%s" % (sys.version_info[0], sys.version_info[1])
        )
    )

    env.SConsignFile(sconsign_filename)


def getArgumentRequired(name):
    """Helper for string options without default value."""
    return scons_arguments[name]


def getArgumentDefaulted(name, default):
    """Helper for string options with default value."""
    return scons_arguments.get(name, default)


def getArgumentInt(option_name, default=None):
    """Small helper for boolean mode flags."""
    if default is None:
        value = scons_arguments[option_name]
    else:
        value = int(scons_arguments.get(option_name, default))

    return value


def getArgumentBool(option_name, default=None):
    """Small helper for boolean mode flags."""
    if default is None:
        value = scons_arguments[option_name]
    else:
        value = scons_arguments.get(option_name, "True" if default else "False")

    return value.lower() in ("yes", "true", "1")


def getArgumentList(option_name, default=None):
    """Small helper for list mode options, default should be command separated str."""
    if default is None:
        value = scons_arguments[option_name]
    else:
        value = scons_arguments.get(option_name, default)

    if value:
        return value.split(",")
    else:
        return []


def enableFlagSettings(env, name, experimental_flags):
    for flag_name in experimental_flags:
        if not flag_name:
            continue

        flag_name = "%s-%s" % (name, flag_name)

        if "=" in flag_name:
            flag_name, value = flag_name.split("=", 1)
        else:
            value = None

        # Allowing for nice names on command line, but using identifiers for C.
        flag_name = flag_name.upper().replace("-", "_").replace(".", "_")

        # Experimental without a value is done as mere define, otherwise
        # the value is passed. spell-checker: ignore cppdefines
        if value:
            env.Append(CPPDEFINES=[("_NUITKA_%s" % flag_name, value)])
        else:
            env.Append(CPPDEFINES=["_NUITKA_%s" % flag_name])


def _prepareFromEnvironmentVar(var_name):
    mingw_mode = False
    zig_mode = False

    # Add environment specified compilers to the PATH variable.
    if var_name in os.environ:
        scons_details_logger.info("%s='%s'" % (var_name, os.environ[var_name]))

        os.environ[var_name] = os.path.normpath(
            os.path.expanduser(os.environ[var_name])
        )

        if os.path.isdir(os.environ[var_name]):
            scons_logger.sysexit(
                "Error, the '%s' variable must point to file, not directory." % var_name
            )

        if os.path.sep in os.environ[var_name]:
            cc_dirname = os.path.dirname(os.environ[var_name])
            if os.path.isdir(cc_dirname):
                addToPATH(None, cc_dirname, prefix=True)

        if os.name == "nt" and isGccName(os.path.basename(os.environ[var_name])):
            scons_details_logger.info(
                "Environment %s seems to be a gcc, enabling mingw_mode." % var_name
            )
            mingw_mode = True

        if isZigName(os.path.basename(os.environ[var_name])):
            scons_details_logger.info(
                "Environment %s seems to be a gcc, enabling zig_mode." % var_name
            )
            zig_mode = True

    return mingw_mode, zig_mode


def _prepareEnvironment(mingw_mode):
    mingw_mode_from_env, zig_mode = _prepareFromEnvironmentVar("CC")
    _prepareFromEnvironmentVar("CXX")

    if mingw_mode_from_env:
        mingw_mode = True

    return mingw_mode, zig_mode


def createEnvironment(
    mingw_mode,
    msvc_version,
    clang_mode,
    clangcl_mode,
    target_arch,
    consider_environ_variables,
    assume_yes_for_downloads,
    source_dir,
):
    # Many settings are directly handled here, getting us a lot of code in here.
    # pylint: disable=too-many-branches,too-many-locals,too-many-statements

    # Zig compiler mode usually from the option --zig.
    zig_exe_path = getArgumentDefaulted("zig_exe_path", None)

    # Prepare environment for compiler detection.
    if consider_environ_variables:
        mingw_mode, zig_mode_env = _prepareEnvironment(mingw_mode=mingw_mode)
    else:
        zig_mode_env = False

    zig_mode = zig_mode_env or zig_exe_path

    from SCons.Script import Environment  # pylint: disable=I0021,import-error

    args = {}

    if msvc_version == "list":
        import SCons.Tool.MSCommon.vc  # pylint: disable=I0021,import-error

        scons_logger.sysexit(
            "Installed MSVC versions are %s."
            % ",".join(repr(v) for v in SCons.Tool.MSCommon.vc.get_installed_vcs()),
        )

    # If we are on Windows, and MinGW is not enforced, lets see if we can
    # find "cl.exe", and if we do, disable automatic scan.
    if (
        os.name == "nt"
        and not mingw_mode
        and not zig_mode
        and msvc_version is None
        and msvc_version != "latest"
        and (getExecutablePath("cl", env=None) is not None)
    ):
        args["MSVC_USE_SCRIPT"] = False

    if mingw_mode or zig_mode or isPosixWindows():
        # Force usage of MinGW64, not using MSVC tools.
        tools = ["mingw"]

        # This code would be running anyway, make it do not thing by monkey patching.
        import SCons.Tool.MSCommon.vc  # pylint: disable=I0021,import-error
        import SCons.Tool.msvc  # pylint: disable=I0021,import-error

        SCons.Tool.MSCommon.vc.msvc_setup_env = lambda *args: None
        SCons.Tool.msvc.msvc_exists = SCons.Tool.MSCommon.msvc_exists = (
            SCons.Tool.MSCommon.vc.msvc_exists
        ) = lambda *args: False
    elif zig_mode:
        tools = ["gcc"]
    else:
        # Everything else should use default, that is MSVC tools, but not MinGW64.
        tools = ["default"]

    env = Environment(
        # We want the outside environment to be passed through.
        ENV=os.environ,
        # Extra tools configuration for scons.
        tools=tools,
        # The shared libraries should not be named "lib...", because CPython
        # requires the filename "module_name.so/pyd" to load it, with no
        # prefix at all, spell-checker: ignore SHLIBPREFIX
        SHLIBPREFIX="",
        # Under windows, specify the target architecture is needed for Scons
        # to pick up MSVC.
        TARGET_ARCH=target_arch,
        # The MSVC version might be fixed by the user.
        MSVC_VERSION=msvc_version if msvc_version != "latest" else None,
        **args
    )

    env.source_dir = source_dir
    env.assume_yes_for_downloads = assume_yes_for_downloads

    if zig_exe_path:
        env["CC"] = zig_exe_path
        env["CXX"] = zig_exe_path
        env["LINK"] = zig_exe_path
        env["CCVERSION"] = None

        # escape backslashes for SCons string interpolation^
        safe_zig_path = zig_exe_path.replace("\\", "\\\\")

        # spell-checker: ignore CCCOM,CFLAGS,CCFLAGS,CCCOMCOM,CXXCOM,CXXFLAGS
        # spell-checker: ignore LINKCOM,LIBDIRFLAGS,LIBFLAGS,SHCCCOM,SHCFLAGS
        # spell-checker: ignore SHCCFLAGS,SHCXXCOM,SHCXXFLAGS,SHLINKCOM,SHLINKFLAGS

        env["CCCOM"] = (
            '"%s" cc -o $TARGET -c $CFLAGS $CCFLAGS $_CCCOMCOM $SOURCES' % safe_zig_path
        )
        env["CXXCOM"] = (
            '"%s" c++ -o $TARGET -c $CXXFLAGS $CCFLAGS $_CCCOMCOM $SOURCES'
            % safe_zig_path
        )
        env["LINKCOM"] = (
            '"%s" cc -o $TARGET $LINKFLAGS $__RPATH $SOURCES $_LIBDIRFLAGS $_LIBFLAGS'
            % safe_zig_path
        )
        env["SHCCCOM"] = (
            '"%s" cc -o $TARGET -c $SHCFLAGS $SHCCFLAGS $_CCCOMCOM $SOURCES'
            % safe_zig_path
        )
        env["SHCXXCOM"] = (
            '"%s" c++ -o $TARGET -c $SHCXXFLAGS $SHCCFLAGS $_CCCOMCOM $SOURCES'
            % safe_zig_path
        )
        env["SHLINKCOM"] = (
            '"%s" cc -o $TARGET $SHLINKFLAGS $__RPATH $SOURCES $_LIBDIRFLAGS $_LIBFLAGS'
            % safe_zig_path
        )

        scons_details_logger.info("Overridden with Zig compiler: %r" % zig_exe_path)

    if consider_environ_variables:
        scons_details_logger.info("Initial CC: %r" % env.get("CC"))
        scons_details_logger.info(
            "Initial CCVERSION: %r" % (env.get("CCVERSION"),),
        )

        if "CC" in os.environ:
            # If the environment variable CC is set, use that.
            env["CC"] = os.environ["CC"]
            env["CCVERSION"] = None

            scons_details_logger.info("Overridden with environment CC: %r" % env["CC"])
        elif clangcl_mode:
            # If possible, add Clang directory from MSVC if available.
            addClangClPathFromMSVC(env=env)
        elif clang_mode and not mingw_mode:
            # If requested by the user, use the clang compiler, overriding what was
            # said in environment.

            env["CC"] = "clang"
            env["CCVERSION"] = None

    # Various flavors could influence builds.
    env.monolithpy = getArgumentBool("monolithpy", False)
    env.debian_python = getArgumentBool("debian_python", False)
    env.fedora_python = getArgumentBool("fedora_python", False)
    env.arch_python = getArgumentBool("arch_python", False)
    env.msys2_mingw_python = getArgumentBool("msys2_mingw_python", False)
    env.anaconda_python = getArgumentBool("anaconda_python", False)
    env.pyenv_python = getArgumentBool("pyenv_python", False)
    env.apple_python = getArgumentBool("apple_python", False)
    env.self_compiled_python_uninstalled = getArgumentBool(
        "self_compiled_python_uninstalled", False
    )

    # No need to pass it from the outside, this cannot use other
    # Pythons for scons.
    env.android_termux_python = isTermuxPython()

    # Non-elf binary, important for linker settings.
    env.noelf_mode = getArgumentBool("noelf_mode", False)

    # Python specific modes have to influence some decisions.
    env.static_libpython = getArgumentDefaulted("static_libpython", "")
    if env.static_libpython:
        assert os.path.exists(env.static_libpython), env.static_libpython

    # Python version we are working on.
    env.python_version_str = getArgumentDefaulted("python_version", None)
    if env.python_version_str is not None:
        env.python_version = tuple(int(d) for d in env.python_version_str.split("."))

        # Do we have a GIL build, or no-GIL (free-threading)
        env.gil_mode = getArgumentBool("gil_mode")

        # Python debug mode: reference count checking, assertions in CPython core.
        env.python_debug = getArgumentBool("python_debug", False)

        # The Python ABI flags to target.
        abiflags = getArgumentDefaulted("abiflags", "")

        if not env.gil_mode and "t" not in abiflags:
            abiflags = "t" + abiflags

        # Python version with ABI included.
        env.python_abi_version = env.python_version_str + abiflags

        # Home of Python to be compiled against, used to find include files and
        # libraries to link against.
        env.python_prefix = getArgumentRequired("python_prefix")

        env.python_prefix_external = getExternalUsePath(env.python_prefix)
    else:
        env.python_version = None

        env.gil_mode = None

        env.python_debug = None

        env.python_abi_version = None

        env.python_prefix = None
        env.python_prefix_external = None

    # Modules count, determines if this is a large compilation.
    env.compiled_module_count = getArgumentInt("compiled_module_count", 0)

    # Target arch for some decisions
    env.target_arch = target_arch

    # Standalone mode
    env.standalone_mode = getArgumentBool("standalone_mode", False)
    if env.standalone_mode:
        env.Append(CPPDEFINES=["_NUITKA_STANDALONE_MODE"])

    # Onefile mode: Create suitable for use in a bootstrap with either dll or
    # exe mode.
    env.onefile_mode = getArgumentBool("onefile_mode", False)
    if env.onefile_mode:
        env.Append(CPPDEFINES=["_NUITKA_ONEFILE_MODE"])

    # Onefile temporary mode: The files are cleaned up after program exit.
    env.onefile_temp_mode = getArgumentBool("onefile_temp_mode", False)
    if env.onefile_temp_mode:
        env.Append(CPPDEFINES=["_NUITKA_ONEFILE_TEMP_BOOL"])

    # Module (or package) mode: Create a Python extension module
    env.module_mode = getArgumentBool("module_mode", False)
    if env.module_mode:
        env.Append(CPPDEFINES=["_NUITKA_MODULE_MODE"])

    # DLL mode: Create a DLL using Python
    env.dll_mode = getArgumentBool("dll_mode", False)
    if env.dll_mode:
        env.Append(CPPDEFINES=["_NUITKA_DLL_MODE"])

    # EXE mode: Create an EXE (using Python for this config)
    env.exe_mode = getArgumentBool("exe_mode", False)
    if env.exe_mode:
        env.Append(CPPDEFINES=["_NUITKA_EXE_MODE"])

    # MacOS bundle: Create an .app on macOS
    env.macos_bundle_mode = getArgumentBool("macos_bundle_mode", False)
    if env.macos_bundle_mode:
        env.Append(CPPDEFINES=["_NUITKA_MACOS_BUNDLE_MODE"])
        env.Append(LINKFLAGS=["-framework", "CoreFoundation"])

    # Announce combination of DLL mode and onefile mode.
    env.onefile_dll_mode = env.onefile_mode and env.dll_mode
    if env.onefile_dll_mode:
        env.Append(CPPDEFINES=["_NUITKA_ONEFILE_DLL_MODE"])

    env.forced_stdout_path = getArgumentDefaulted("forced_stdout_path", None)
    env.forced_stderr_path = getArgumentDefaulted("forced_stderr_path", None)

    env.build_definitions = {}

    return env


def decodeData(data):
    """Our own decode tries to workaround MSVC misbehavior."""
    try:
        return data.decode(sys.__stdout__.encoding)
    except UnicodeDecodeError:
        import locale

        try:
            return data.decode(locale.getpreferredencoding())
        except UnicodeDecodeError:
            return data.decode("utf8", "backslashreplace")


def getExecutablePath(filename, env):
    """Find an execute in either normal PATH, or Scons detected PATH."""

    if os.path.exists(filename):
        return filename

    # Variable substitution from environment is needed, because this can contain
    # "$CC" which should be looked up too.
    while filename.startswith("$"):
        filename = env[filename[1:]]

    # Append ".exe" suffix  on Windows if not already present.
    if os.name == "nt" and not filename.lower().endswith(".exe"):
        filename += ".exe"

    # Either look at the initial "PATH" as given from the outside or look at the
    # current environment.
    if env is None:
        search_path = os.environ["PATH"]
    else:
        search_path = env._dict["ENV"]["PATH"]  # pylint: disable=protected-access

    # Now check in each path element, much like the shell will.
    path_elements = search_path.split(os.pathsep)

    for path_element in path_elements:
        path_element = path_element.strip('"')

        full = getNormalizedPathJoin(path_element, filename)

        if os.path.isfile(full):
            return full

    return None


def changeKeyboardInterruptToErrorExit():
    def signalHandler(
        signal, frame
    ):  # pylint: disable=redefined-outer-name,unused-argument
        sys.exit(2)

    signal.signal(signal.SIGINT, signalHandler)


def setEnvironmentVariable(env, key, value):
    if value is None:
        del os.environ[key]
    elif key in os.environ:
        os.environ[key] = value

    if env is not None:
        # pylint: disable=protected-access
        if value is None:
            del env._dict["ENV"][key]
        else:
            env._dict["ENV"][key] = value


def addToPATH(env, dirname, prefix):
    # Otherwise subprocess will complain in Python2
    if str is bytes and type(dirname) is unicode:
        dirname = dirname.encode("utf8")

    path_value = os.environ["PATH"].split(os.pathsep)

    if prefix:
        path_value.insert(0, dirname)
    else:
        path_value.append(dirname)

    setEnvironmentVariable(env, "PATH", os.pathsep.join(path_value))


def writeSconsReport(env, target):
    with openTextFile(
        _getSconsReportFilename(env.source_dir), "w", encoding="utf8"
    ) as report_file:
        # We are friends to get at this debug info, pylint: disable=protected-access
        for key, value in sorted(env._dict.items()):
            if type(value) is list and all(isinstance(v, basestring) for v in value):
                value = repr(value)

            if not isinstance(value, basestring):
                continue

            if key.startswith(("_", "CONFIGURE")):
                continue

            # Ignore problematic and useless values
            # spell-checker: ignore MSVSSCONS,IDLSUFFIXES,DSUFFIXES
            if key in ("MSVSSCONS", "BUILD_DIR", "IDLSUFFIXES", "DSUFFIXES"):
                continue

            # TODO: For these kinds of prints, maybe have our own method of doing them
            # rather than print, or maybe just json or something similar.
            print(key + "=" + value, file=report_file)

        print("gcc_mode=%s" % env.gcc_mode, file=report_file)
        print("clang_mode=%s" % env.clang_mode, file=report_file)
        print("msvc_mode=%s" % env.msvc_mode, file=report_file)
        print("mingw_mode=%s" % env.mingw_mode, file=report_file)
        print("clangcl_mode=%s" % env.clangcl_mode, file=report_file)

        print("cpp_flags=%s" % (env.cpp_flags or ""), file=report_file)
        print("c_flags=%s" % (env.c_flags or ""), file=report_file)
        print("cc_flags=%s" % (env.cc_flags or ""), file=report_file)
        print("cxx_flags=%s" % (env.cxx_flags or ""), file=report_file)
        print("ld_flags=%s" % (env.ld_flags or ""), file=report_file)

        print("PATH=%s" % os.environ["PATH"], file=report_file)
        print("TARGET=%s" % getNormalizedPath(target[0].abspath), file=report_file)


_checked_msvc_language_pack = False


def reportSconsUnexpectedOutput(env, cmdline, stdout, stderr):
    # Singleton, pylint: disable=global-statement
    global _checked_msvc_language_pack

    if not _checked_msvc_language_pack and env.msvc_mode and (stderr or stdout):
        _checked_msvc_language_pack = True

        cl_path = getExecutablePath(env.the_compiler, env=env)
        bin_dir = os.path.dirname(cl_path)
        eng_dir = os.path.join(bin_dir, "1033")

        if not os.path.exists(eng_dir):
            scons_logger.warning(
                """\
Support language of Nuitka is English. Please install the English \
language pack for Visual Studio in the installer. There is a \
section for that."""
            )

    if env.warn_error_mode and stderr is not None:

        if (
            # gcc does this: "all warnings being treated as errors"
            "all warnings being treated as errors" in stderr
            or
            # clang does this, spell-checker: ignore Werror
            "[-Werror," in stderr
            or
            # macOS clang does this: "warnings generated"
            "warnings generated" in stderr
        ):
            scons_logger.sysexit(
                "Error, C warnings occurred with '--debug' or '--debug-c-warnings', use '--no-debug-c-warnings' or fix them."
            )

    with withFileLock("writing scons error report"):
        file_handle, pickler = openPickleFile(
            _getSconsErrorReportFilename(env.source_dir), "ab", protocol=2
        )
        pickler.dump((cmdline, stdout, stderr))
        file_handle.close()


_scons_reports = {}
_scons_error_reports = {}


def flushSconsReports():
    _scons_reports.clear()
    _scons_error_reports.clear()


def _getSconsReportFilename(source_dir):
    return getNormalizedPathJoin(source_dir, "scons-report.txt")


def _getSconsErrorReportFilename(source_dir):
    return getNormalizedPathJoin(source_dir, "scons-error-report.txt")


def readSconsReport(source_dir):
    if source_dir not in _scons_reports:
        scons_report = OrderedDict()

        for line in getFileContentByLine(
            _getSconsReportFilename(source_dir), encoding="utf8"
        ):
            if "=" not in line:
                continue

            key, value = line.strip().split("=", 1)

            scons_report[key] = value

        _scons_reports[source_dir] = scons_report

    return _scons_reports[source_dir]


def getSconsReportValue(source_dir, key, default=Ellipsis):
    try:
        return readSconsReport(source_dir).get(key)
    except FileNotFoundError:
        if default is not Ellipsis:
            return default

        raise


def readSconsErrorReport(source_dir):
    if source_dir not in _scons_error_reports:
        scons_error_report = OrderedDict()

        scons_error_report_filename = _getSconsErrorReportFilename(source_dir)
        if os.path.exists(scons_error_report_filename):
            file_handle = openTextFile(scons_error_report_filename, "rb")
            try:
                while True:
                    try:
                        cmd, stdout, stderr = pickle.load(file_handle)
                    except EOFError:
                        break

                    if type(cmd) in (tuple, list):
                        cmd = " ".join(cmd)

                    if cmd not in scons_error_report:
                        scons_error_report[cmd] = ["", ""]

                    if type(stdout) in (tuple, list):
                        stdout = "\n".join(stdout)

                    if type(stderr) in (tuple, list):
                        stderr = "\n".join(stderr)

                    if stdout:
                        stdout = stdout.replace("\n\r", "\n")
                        scons_error_report[cmd][0] += stdout
                    if stderr:
                        stderr = stderr.replace("\n\r", "\n")
                        scons_error_report[cmd][1] += stderr

            finally:
                file_handle.close()

        _scons_error_reports[source_dir] = scons_error_report

    return _scons_error_reports[source_dir]


def addClangClPathFromMSVC(env):
    cl_exe = getExecutablePath("cl", env=env)

    if cl_exe is None:
        scons_logger.sysexit(
            "Error, Visual Studio required for using ClangCL on Windows."
        )

    clang_dir = getNormalizedPathJoin(cl_exe[: cl_exe.lower().rfind("msvc")], "Llvm")

    if (
        getCompilerArch(
            mingw_mode=False, msvc_mode=True, the_cc_name="cl.exe", compiler_path=cl_exe
        )
        == "pei-x86-64"
    ):
        clang_dir = getNormalizedPathJoin(clang_dir, "x64", "bin")
    else:
        clang_dir = getNormalizedPathJoin(clang_dir, "bin")

    if not os.path.exists(clang_dir):
        scons_details_logger.sysexit(
            "Visual Studio has no Clang component found at '%s'." % clang_dir
        )

    scons_details_logger.info(
        "Adding Visual Studio directory '%s' for Clang to PATH." % clang_dir
    )

    addToPATH(env, clang_dir, prefix=True)

    clangcl_path = getExecutablePath("clang-cl", env=env)

    if clangcl_path is None:
        scons_details_logger.sysexit(
            "Visual Studio has no Clang component found at '%s'." % clang_dir
        )

    env["CC"] = "clang-cl"
    env["LINK"] = "lld-link"

    # Version information is outdated now, spell-checker: ignore ccversion
    env["CCVERSION"] = None


def isGccName(cc_name):
    cc_name = os.path.normcase(os.path.basename(cc_name))

    return (
        "gcc" in cc_name
        or "g++" in cc_name
        or "gnu-cc" in cc_name
        or "gnu-gcc" in cc_name
    )


def isClangName(cc_name):
    cc_name = os.path.normcase(os.path.basename(cc_name))

    return ("clang" in cc_name and "-cl" not in cc_name) or isZigName(cc_name)


def isZigName(cc_name):
    cc_name = os.path.normcase(os.path.basename(cc_name))

    return "zig" in cc_name


def scanSourceDir(env, dirname, plugins):
    if not os.path.exists(dirname):
        return

    # If we use C11 capable compiler, all good. Otherwise use C++, which Scons
    # needs to derive from filenames, so make copies (or links) with a different
    # name.
    added_path = False

    filenames = sorted(os.listdir(dirname))

    for filename_base in filenames:
        # Only C files are of interest here.
        if not hasFilenameExtension(filename_base, (".c", ".cpp")):
            continue

        # If we have a C file, but a C++ file exists too, use that.
        if filename_base.endswith(".c") and (filename_base[:-2] + ".cpp") in filenames:
            continue

        if filename_base.endswith(".h") and plugins and not added_path:
            # Adding path for source paths on the fly, spell-checker: ignore cpppath
            env.Append(CPPPATH=[dirname])
            added_path = True

        filename = getNormalizedPathJoin(dirname, filename_base)

        target_filename = filename

        if isWin32Windows() and not isFilesystemEncodable(filename_base):
            target_filename = getWindowsShortPathName(target_filename)

            # Avoid ".C" suffixes, that MinGW64 wouldn't recognize.
            target_filename = changeFilenameExtension(
                target_filename, getFilenameExtension(target_filename).lower()
            )

        # We pretend to use C++ if no C11 compiler is present.
        if env.c11_mode:
            yield target_filename
        else:
            if hasFilenameExtension(filename, ".c"):
                target_filename += "pp"  # .cpp" suffix then

                os.rename(filename, target_filename)

            yield target_filename


def makeCLiteral(value):
    value = value.replace("\\", r"\\")
    value = value.replace('"', r"\"")

    return '"' + value + '"'


def createDefinitionsFile(source_dir, filename, definitions):
    for env_name in os.environ["_NUITKA_BUILD_DEFINITIONS_CATALOG"].split(","):
        definitions[env_name] = os.environ[env_name]

    build_definitions_filename = getNormalizedPathJoin(source_dir, filename)

    with openTextFile(build_definitions_filename, "w", encoding="utf8") as f:
        for key, value in sorted(definitions.items()):
            if key == "_NUITKA_BUILD_DEFINITIONS_CATALOG":
                continue

            if type(value) is int or key.endswith(("_BOOL", "_INT")):
                if type(value) is bool:
                    value = int(value)
                f.write("#define %s %s\n" % (key, value))
            elif type(value) in (str, unicode) and key.endswith("_WIDE_STRING"):
                f.write("#define %s L%s\n" % (key, makeCLiteral(value)))
            else:
                f.write("#define %s %s\n" % (key, makeCLiteral(value)))


def getMsvcVersionString(env):
    import SCons.Tool.MSCommon.vc  # pylint: disable=I0021,import-error

    return SCons.Tool.MSCommon.vc.get_default_version(env)


def getMsvcVersion(env):
    value = getMsvcVersionString(env)

    # TODO: Workaround for prompt being used.
    if value is None:
        value = os.getenv("VCToolsVersion", "14.3").rsplit(".", 1)[0]

    value = value.replace("exp", "")
    return tuple((int(d) for d in value.split(".")))


def _getBinaryArch(binary, mingw_mode):
    if isLinux() or mingw_mode:
        assert os.path.exists(binary), binary

        # Binutils binary name, spell-checker: ignore objdump,binutils
        command = ["objdump", "-f", binary]

        try:
            data, _err, rv = executeProcess(command)
        except OSError:
            command[0] = "llvm-objdump"

            try:
                data, _err, rv = executeProcess(command)
            except OSError:
                return None

        if rv != 0:
            return None

        if str is not bytes:
            data = decodeData(data)

        found = None

        for line in data.splitlines():
            if " file format " in line:
                found = line.split(" file format ")[-1]
            if "\tfile format " in line:
                found = line.split("\tfile format ")[-1]

        if os.name == "nt" and found == "coff-x86-64":
            found = "pei-x86-64"

        return found
    else:
        # TODO: Missing for macOS, FreeBSD, other Linux
        return None


_linker_arch_determined = False
_linker_arch = None


def getLinkerArch(target_arch, mingw_mode):
    # Singleton, pylint: disable=global-statement
    global _linker_arch_determined, _linker_arch

    if not _linker_arch_determined:
        if os.name == "nt":
            if target_arch == "x86_64":
                _linker_arch = "pei-x86-64"
            elif target_arch == "arm64":
                _linker_arch = "pei-arm64"
            else:
                _linker_arch = "pei-i386"
        else:
            _linker_arch = _getBinaryArch(
                binary=os.environ["NUITKA_PYTHON_EXE_PATH"], mingw_mode=mingw_mode
            )

        _linker_arch_determined = True

    return _linker_arch


_compiler_arch = {}


def getCompilerArch(mingw_mode, msvc_mode, the_cc_name, compiler_path):
    assert not mingw_mode or not msvc_mode

    if compiler_path not in _compiler_arch:
        if mingw_mode:
            _compiler_arch[compiler_path] = _getBinaryArch(
                binary=compiler_path, mingw_mode=mingw_mode
            )
        elif msvc_mode:
            cmdline = [compiler_path]

            if "-cl" in the_cc_name:
                cmdline.append("--version")

            # The cl.exe without further args will give error
            stdout, stderr, _rv = executeProcess(
                command=cmdline,
            )

            # The MSVC will output on error, while clang outputs in stdout and they
            # use different names for arches.
            if b"x64" in stderr or b"x86_64" in stdout:
                _compiler_arch[compiler_path] = "pei-x86-64"
            elif b"x86" in stderr or b"i686" in stdout:
                _compiler_arch[compiler_path] = "pei-i386"
            elif b"ARM64" in stderr:
                # TODO: The ARM64 output for Clang is not known yet.
                _compiler_arch[compiler_path] = "pei-arm64"
            else:
                assert False, (stdout, stderr)
        else:
            assert False, compiler_path

    return _compiler_arch[compiler_path]


def decideArchMismatch(target_arch, the_cc_name, compiler_path):
    mingw_mode = isGccName(the_cc_name) or isClangName(the_cc_name)
    msvc_mode = not mingw_mode

    linker_arch = getLinkerArch(target_arch=target_arch, mingw_mode=mingw_mode)
    compiler_arch = getCompilerArch(
        mingw_mode=mingw_mode,
        msvc_mode=msvc_mode,
        the_cc_name=the_cc_name,
        compiler_path=compiler_path,
    )

    return linker_arch != compiler_arch, linker_arch, compiler_arch


def raiseNoCompilerFoundErrorExit():
    if os.name == "nt":
        scons_logger.sysexit(
            """\
Error, cannot locate suitable C compiler. You have the following options:

a) If a suitable Visual Studio version is installed (check above trace
   outputs for rejection messages), it will be located automatically via
   registry. But not if you activate the wrong prompt, so don't do that
   unless you know what you are doing.

b) Using "--mingw64" forces Nuitka download MinGW64 for you, but only for
   Python version 3.12 and below, newer versions cannot use it. Note: MinGW64
   is the project name, it does *not* mean 64 bits, just a gcc with better
   Windows compatibility, it is available for 32 and 64 bits. Cygwin based
   gcc e.g. do not work.

c) Using "--zig" forces Nuitka download and use Zig for C compilation, but
   this only works with 64 bit Python versions.
"""
        )
    else:
        scons_logger.sysexit("Error, cannot locate suitable C compiler.")


def addBinaryBlobSection(env, blob_filename, section_name):
    # spell-checker: ignore linkflags, sectcreate

    if isMacOS():
        env.Append(
            LINKFLAGS=[
                "-Wl,-sectcreate,%(section_name)s,%(section_name)s,%(blob_filename)s"
                % {
                    "section_name": section_name,
                    "blob_filename": blob_filename,
                }
            ]
        )
    else:
        assert False


def makeResultPathFileSystemEncodable(env, result_exe):
    deleteFile(result_exe, must_exist=False)

    if os.name == "nt" and not isFilesystemEncodable(result_exe):
        result_exe = getNormalizedPathJoin(
            os.path.dirname(result_exe),
            "_nuitka_temp.pyd" if env.module_mode else "_nuitka_temp.exe",
        )

        if not isFilesystemEncodable(result_exe):
            result_exe = getNormalizedPath(os.path.relpath(result_exe))

            deleteFile(result_exe, must_exist=False)

    return result_exe


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.gnu.org/licenses/agpl.txt
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2025, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Exception handling."""

from nuitka.PythonVersions import python_version

from .CodeHelpers import (
    generateExpressionCode,
    withObjectCodeTemporaryAssignment,
)
from .ErrorCodes import getErrorExitCode
from .templates.CodeTemplatesExceptions import (
    template_publish_exception_to_handler,
)


def getExceptionIdentifier(exception_type):
    assert "PyExc" not in exception_type, exception_type

    if exception_type == "NotImplemented":
        return "Py_NotImplemented"

    return "PyExc_%s" % exception_type


def generateExceptionRefCode(to_name, expression, emit, context):
    exception_type = expression.getExceptionName()

    with withObjectCodeTemporaryAssignment(
        to_name, "exception_name", expression, emit, context
    ) as value_name:
        emit("%s = %s;" % (value_name, getExceptionIdentifier(exception_type)))


def getTracebackMakingIdentifier(context, lineno_name):
    frame_handle = context.getFrameHandle()
    assert frame_handle is not None

    return "MAKE_TRACEBACK(%s, %s)" % (frame_handle, lineno_name)


def generateExceptionCaughtTypeCode(to_name, expression, emit, context):
    (
        keeper_exception_state_name,
        _keeper_lineno,
    ) = context.getExceptionKeeperVariables()

    with withObjectCodeTemporaryAssignment(
        to_name, "exception_caught_type", expression, emit, context
    ) as value_name:
        if keeper_exception_state_name is None:
            emit("%s = EXC_TYPE(tstate);" % (value_name,))
        else:
            if python_version < 0x3C0:
                emit(
                    "%s = %s.exception_type;"
                    % (value_name, keeper_exception_state_name)
                )
            else:
                emit(
                    "%s = %s.exception_value;"
                    % (value_name, keeper_exception_state_name)
                )


def generateExceptionCaughtValueCode(to_name, expression, emit, context):
    (
        keeper_exception_state_name,
        _keeper_lineno,
    ) = context.getExceptionKeeperVariables()

    with withObjectCodeTemporaryAssignment(
        to_name, "exception_caught_value", expression, emit, context
    ) as value_name:
        if keeper_exception_state_name is None:
            emit("%s = EXC_VALUE(tstate);" % value_name)
        elif python_version >= 0x270:
            emit("%s = %s.exception_value;" % (value_name, keeper_exception_state_name))
        else:
            # For Python2.6, value can be NULL.
            emit(
                "%s = %s.exception_value ? %s.exception_value : Py_None;"
                % (
                    value_name,
                    keeper_exception_state_name,
                    keeper_exception_state_name,
                )
            )

        emit("CHECK_OBJECT(%s); " % value_name)


def generateExceptionCaughtTracebackCode(to_name, expression, emit, context):
    (
        keeper_exception_state_name,
        keeper_lineno,
    ) = context.getExceptionKeeperVariables()

    with withObjectCodeTemporaryAssignment(
        to_name, "exception_caught_tb", expression, emit, context
    ) as value_name:
        if keeper_exception_state_name is None:
            if python_version < 0x3B0:
                emit("%s = (PyObject *)EXC_TRACEBACK(tstate);" % (value_name,))
            else:
                emit(
                    """\
%(value_name)s = (PyObject *)GET_EXCEPTION_TRACEBACK(EXC_VALUE(tstate));
if (%(value_name)s == NULL) {
    %(value_name)s = Py_None;
}"""
                    % {"value_name": value_name}
                )

        else:
            emit(
                """\
if (%(keeper_exception_state_name)s.exception_tb != NULL) {
    %(to_name)s = (PyObject *)%(keeper_exception_state_name)s.exception_tb;
    Py_INCREF(%(to_name)s);
} else {
    %(to_name)s = (PyObject *)%(tb_making)s;
}
"""
                % {
                    "to_name": value_name,
                    "keeper_exception_state_name": keeper_exception_state_name,
                    "tb_making": getTracebackMakingIdentifier(
                        context=context, lineno_name=keeper_lineno
                    ),
                }
            )

            context.addCleanupTempName(value_name)


def getExceptionUnpublishedReleaseCode(emit, context):
    (
        keeper_exception_state_name,
        _keeper_lineno,
    ) = context.getExceptionKeeperVariables()

    if keeper_exception_state_name is not None:
        emit("RELEASE_ERROR_OCCURRED_STATE(&%s);" % keeper_exception_state_name)


def generateExceptionPublishCode(statement, emit, context):
    # This statement has no attributes really, pylint: disable=unused-argument

    # Current variables cannot be used anymore now.
    (
        keeper_exception_state_name,
        keeper_lineno,
    ) = context.setExceptionKeeperVariables((None, None))

    emit(
        template_publish_exception_to_handler
        % {
            "tb_making": getTracebackMakingIdentifier(
                context=context, lineno_name=keeper_lineno
            ),
            "keeper_exception_state_name": keeper_exception_state_name,
            "keeper_lineno": keeper_lineno,
            "frame_identifier": context.getFrameHandle(),
        }
    )

    # TODO: Make this one thing for performance with thread state shared, also for less code,
    # then we should not make it in header anymore. Might be more scalable too.
    emit("PUBLISH_CURRENT_EXCEPTION(tstate, &%s);" % keeper_exception_state_name)


def _attachExceptionAttributeCode(
    to_name,
    attribute_expression,
    c_type_name,
    c_attribute_name,
    base_name_str,
    emit,
    context,
):
    if attribute_expression is not None:
        from .PythonAPICodes import getReferenceExportCode

        exception_attribute_name = context.allocateTempName(
            base_name_str + "_" + c_attribute_name
        )

        generateExpressionCode(
            to_name=exception_attribute_name,
            expression=attribute_expression,
            emit=emit,
            context=context,
            allow_none=True,
        )

        getReferenceExportCode(exception_attribute_name, emit, context)
        if context.needsCleanup(exception_attribute_name):
            context.removeCleanupTempName(exception_attribute_name)

        emit(
            "((%s *)%s)->%s = %s;"
            % (c_type_name, to_name, c_attribute_name, exception_attribute_name)
        )


def _generateBuiltinMakeExceptionCode(to_name, expression, for_raise, emit, context):
    from .CallCodes import getCallCodeNoArgs, getCallCodePosArgsQuick

    exception_arg_names = []

    for exception_arg in expression.subnode_args:
        exception_arg_name = context.allocateTempName("make_exception_arg")

        generateExpressionCode(
            to_name=exception_arg_name,
            expression=exception_arg,
            emit=emit,
            context=context,
        )

        exception_arg_names.append(exception_arg_name)

    exception_type = expression.getExceptionName()

    with withObjectCodeTemporaryAssignment(
        to_name, "exception_made", expression, emit, context
    ) as value_name:
        exception_name = getExceptionIdentifier(exception_type)

        if len(exception_arg_names) == 1 and for_raise:
            emit(
                "%s = MAKE_EXCEPTION_WITH_VALUE(tstate, %s, %s);"
                % (value_name, exception_name, exception_arg_names[0])
            )

            getErrorExitCode(
                check_name=value_name,
                release_names=exception_arg_names,
                needs_check=expression.mayRaiseExceptionOperation(),
                emit=emit,
                context=context,
            )

            context.addCleanupTempName(value_name)
        elif exception_arg_names:
            getCallCodePosArgsQuick(
                to_name=value_name,
                called_name=exception_name,
                expression=expression,
                arg_names=exception_arg_names,
                emit=emit,
                context=context,
            )

        else:
            getCallCodeNoArgs(
                to_name=value_name,
                called_name=exception_name,
                expression=expression,
                emit=emit,
                context=context,
            )

        if (
            expression.isExpressionBuiltinMakeExceptionImportError()
            or expression.isExpressionBuiltinMakeExceptionModuleNotFoundError()
        ):
            _attachExceptionAttributeCode(
                to_name=to_name,
                attribute_expression=expression.subnode_name,
                base_name_str="exception_import_error",
                c_type_name="PyImportErrorObject",
                c_attribute_name="name",
                emit=emit,
                context=context,
            )

            _attachExceptionAttributeCode(
                to_name=to_name,
                attribute_expression=expression.subnode_path,
                base_name_str="exception_import_error",
                c_type_name="PyImportErrorObject",
                c_attribute_name="path",
                emit=emit,
                context=context,
            )

        elif expression.isExpressionBuiltinMakeExceptionAttributeError():
            _attachExceptionAttributeCode(
                to_name=to_name,
                attribute_expression=expression.subnode_name,
                base_name_str="exception_import_error",
                c_type_name="PyAttributeErrorObject",
                c_attribute_name="name",
                emit=emit,
                context=context,
            )

            _attachExceptionAttributeCode(
                to_name=to_name,
                attribute_expression=expression.subnode_obj,
                base_name_str="exception_import_error",
                c_type_name="PyAttributeErrorObject",
                c_attribute_name="obj",
                emit=emit,
                context=context,
            )


def generateBuiltinMakeExceptionCode(to_name, expression, emit, context):
    _generateBuiltinMakeExceptionCode(
        to_name=to_name,
        expression=expression,
        for_raise=expression.for_raise,
        emit=emit,
        context=context,
    )


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.gnu.org/licenses/agpl.txt
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2025, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""This test uses the meta data of metadata lookups that wouldn't work at runtime.

But since Nuitka resolves them at compile time, no issue should happen, unless
the module used in the example is not installed.
"""

# nuitka-project: --mode=standalone

import sys

# nuitka-skip-unless-imports: lxml

try:
    from pkg_resources import get_distribution, require
except ImportError:
    print("No pkg_resources found")
else:
    x = require("lxml")
    print(x[0].version)
    print(x[0].parsed_version)
    __version__ = get_distribution("lxml").version
    print("pkg_resources gives version", __version__)
    __version__ = get_distribution("lxml").parsed_version
    print("pkg_resources gives parsed version", __version__)


try:
    from importlib.metadata import version
except ImportError:
    print("Old Python", sys.version_info)
else:
    __version__ = version("lxml")
    print("Stdlib importlib.metadata gives version", __version__)


try:
    from importlib_metadata import version
except ImportError:
    print("Backport importlib_metadata is not installed.")
else:
    __version__ = version("lxml")
    print("Backport importlib_metadata gives version", __version__)

#     Python test originally created or extracted from other peoples work. The
#     parts from me are licensed as below. It is at least Free Software where
#     it's copied from other people. In these cases, that will normally be
#     indicated.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



# MIT License
#
# Copyright The SCons Foundation

"""SCons file locking functions.

Simple-minded filesystem-based locking. Provides a context manager
which acquires a lock (or at least, permission) on entry and
releases it on exit.

Usage::

    from SCons.Util.filelock import FileLock

    with FileLock("myfile.txt", writer=True) as lock:
        print(f"Lock on {lock.file} acquired.")
        # work with the file as it is now locked
"""

# TODO: things to consider.
#   Is raising an exception the right thing for failing to get lock?
#   Is a filesystem lockfile scheme sufficient for our needs?
#   - or is it better to put locks on the actual file (fcntl/windows-based)?
#   ... Is that even viable in the case of a remote (network) file?
#   Is this safe enough? Or do we risk dangling lockfiles?
#   Permission issues in case of multi-user. This *should* be okay,
#     the cache usually goes in user's homedir, plus you already have
#     enough rights for the lockfile if the dir lets you create the cache.
#   Need a forced break-lock method?
#   The lock attributes could probably be made opaque. Showed one visible
#     in the example above, but not sure the benefit of that.

from __future__ import annotations

import os
import time


class SConsLockFailure(Exception):
    """Lock failure exception."""


class FileLock:
    """Lock a file using a lockfile.

    Basic locking for when multiple processes may hit an externally
    shared resource that cannot depend on locking within a single SCons
    process. SCons does not have a lot of those, but caches come to mind.

    Cross-platform safe, does not use any OS-specific features.  Provides
    context manager support, or can be called with :meth:`acquire_lock`
    and :meth:`release_lock`.

    Lock can be a write lock, which is held until released, or a read
    lock, which releases immediately upon aquisition - we want to not
    read a file which somebody else may be writing, but not create the
    writers starvation problem of the classic readers/writers lock.

    TODO: Should default timeout be None (non-blocking), or 0 (block forever),
       or some arbitrary number?

    Arguments:
       file: name of file to lock. Only used to build the lockfile name.
       timeout: optional time (sec) to give up trying.
          If ``None``, quit now if we failed to get the lock (non-blocking).
          If 0, block forever (well, a long time).
       delay: optional delay between tries [default 0.05s]
       writer: if True, obtain the lock for safe writing. If False (default),
          just wait till the lock is available, give it back right away.

    Raises:
        SConsLockFailure: if the operation "timed out", including the
          non-blocking mode.
    """

    def __init__(
        self,
        file,
        timeout = None,
        delay = 0.05,
        writer = False,
    ):
        if timeout is not None and delay is None:
            raise ValueError("delay cannot be None if timeout is None.")
        # It isn't completely obvious where to put the lockfile.
        # This scheme depends on diffrent processes using the same path
        # to the lockfile, since the lockfile is the magic resource,
        # not the file itself. getcwd() is no good for testcases, each of
        # which run in a unique test directory. tempfile is no good,
        # as those are (intentionally) unique per process.
        # Our simple first guess is just put it where the file is.
        self.file = file
        self.lockfile = f"{file}.lock"
        self.lock: int | None = None
        self.timeout = 999999 if timeout == 0 else timeout
        self.delay = 0.0 if delay is None else delay
        self.writer = writer

    def acquire_lock(self):
        """Acquire the lock, if possible.

        If the lock is in use, check again every *delay* seconds.
        Continue until lock acquired or *timeout* expires.
        """
        start_time = time.perf_counter()
        while True:
            try:
                self.lock = os.open(self.lockfile, os.O_CREAT|os.O_EXCL|os.O_RDWR)
            except (FileExistsError, PermissionError) as exc:
                if self.timeout is None:
                    raise SConsLockFailure(
                        f"Could not acquire lock on {self.file!r}"
                    ) from exc
                if (time.perf_counter() - start_time) > self.timeout:
                    raise SConsLockFailure(
                        f"Timeout waiting for lock on {self.file!r}."
                    ) from exc
                time.sleep(self.delay)
            else:
                if not self.writer:
                    # reader: waits to get lock, but doesn't hold it
                    self.release_lock()
                break

    def release_lock(self):
        """Release the lock by deleting the lockfile."""
        if self.lock:
            os.close(self.lock)
            os.unlink(self.lockfile)
            self.lock = None

    def __enter__(self):
        """Context manager entry: acquire lock if not holding."""
        if not self.lock:
            self.acquire_lock()
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        """Context manager exit: release lock if holding."""
        if self.lock:
            self.release_lock()

    def __repr__(self):
        """Nicer display if someone repr's the lock class."""
        return (
            f"{self.__class__.__name__}("
            f"file={self.file!r}, "
            f"timeout={self.timeout!r}, "
            f"delay={self.delay!r}, "
            f"writer={self.writer!r})"
        )
